"""Wrapper for the sampling point operator."""
from typing import Any, Collection, Dict, List, Sequence, cast

from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from .constants import Spot
from .result_definition import FailureMeasureEnum
from .sampling_point_types import (
    FAILURE_MODE_NAMES_TO_ACP,
    FailureResult,
    SamplingPoint,
    SamplingPointFigure,
)


def get_analysis_plies_from_sp(results: Any) -> Sequence[Any]:
    """List of analysis plies from the bottom to the top.

    This attribute returns a list of ply data, such as angle, thickness and material name,
    as a dictionary.
    """
    if len(results) == 0:
        raise RuntimeError("Cannot extract Analysis Plies from the Sampling Point result.")

    raw_data = cast(Sequence[Any], results[0]["layup"]["analysis_plies"])
    if len(raw_data) == 0:
        raise RuntimeError("No plies are found for the selected element.")

    types = {
        "angle": float,
        "global_ply_number": int,
        "id": str,
        "is_core": bool,
        "material": str,
        "thickness": float,
    }

    plies = []
    for raw_ply in raw_data:
        plies.append({k: types[k](v) for k, v in raw_ply.items()})

    return plies


def get_data_from_sp_results(*args: Any, results: Any) -> npt.NDArray[np.float64]:
    """Extract data from the result dict.

    The result object of the sampling point is a (nested) JSON dict. The args are the keys
    to dive into the results and extract the according data.
    For example: get_data_from_sp_results("results", "stresses", "s1", results=sp.results)
    returns the stresses in the material direction.
    """

    def _next(outer_data: Any, comp: str) -> Any:
        if comp in outer_data.keys():
            return outer_data[comp]

        raise RuntimeError(f"Cannot extract result {comp} from Sampling Point results.")

    if not results or len(results) == 0:
        raise RuntimeError(f"Cannot extract result {args} from Sampling Point result.")

    data = results[0]
    for key in args:
        data = _next(data, key)
    return np.array(data)


def get_indices_from_sp(
    interface_indices: Dict[Spot, int],
    number_of_plies: int,
    spots_per_ply: int,
    spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
) -> Sequence[int]:
    """Get the indices of the selected spots (interfaces) for each ply."""
    ply_wise_indices = [interface_indices[v] for v in spots]
    ply_wise_indices.sort()
    indices = []
    for ply_index in range(0, number_of_plies):
        indices.extend([ply_index * spots_per_ply + index for index in ply_wise_indices])

    return indices


def get_offsets_by_spots_from_sp(
    sampling_point: SamplingPoint,
    spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
    core_scale_factor: float = 1.0,
) -> npt.NDArray[np.float64]:
    """Access the y coordinates of the selected spots (interfaces) for each ply."""
    offsets = sampling_point.offsets
    indices = sampling_point.get_indices(spots)

    if core_scale_factor == 1.0:
        return cast(npt.NDArray[np.float64], offsets[indices])

    spots_per_ply = sampling_point.spots_per_ply

    thicknesses = []
    if not sampling_point.analysis_plies:
        raise RuntimeError("No analysis plies are found in the selected element.")

    for index, ply in enumerate(sampling_point.analysis_plies):
        is_core = ply["is_core"]
        # get thickness from the offsets
        th = offsets[(index + 1) * spots_per_ply - 1] - offsets[index * spots_per_ply]
        if is_core:
            th *= core_scale_factor

        thicknesses.append(th)

    for index, ply in enumerate(sampling_point.analysis_plies):
        # spots per ply is always 2 or 3
        step = thicknesses[index] / (spots_per_ply - 1)
        top_of_previous_ply = offsets[index * spots_per_ply - 1] if index > 0 else offsets[0]
        for i in range(0, spots_per_ply):
            offsets[index * spots_per_ply + i] = top_of_previous_ply + step * i

    return cast(npt.NDArray[np.float64], offsets[indices])


def get_ply_wise_critical_failures_from_sp(
    sampling_point: SamplingPoint,
) -> List[FailureResult]:
    """Get the critical failure value and modes per ply."""
    num_plies = sampling_point.number_of_plies

    irfs = sampling_point.inverse_reserve_factor.reshape(num_plies, sampling_point.spots_per_ply)
    mos = sampling_point.margin_of_safety.reshape(num_plies, sampling_point.spots_per_ply)
    rfs = sampling_point.reserve_factor.reshape(num_plies, sampling_point.spots_per_ply)
    failure_models = np.array(sampling_point.failure_modes).reshape(
        num_plies, sampling_point.spots_per_ply
    )

    critical_indices = irfs.argmax(1)

    result = []

    for ply_index, crit_index in enumerate(critical_indices):
        result.append(
            FailureResult(
                failure_models[ply_index][crit_index],
                irfs[ply_index][crit_index],
                rfs[ply_index][crit_index],
                mos[ply_index][crit_index],
            )
        )

    return result


def get_polar_plot_from_sp(
    sampling_point: SamplingPoint, components: Sequence[str] = ("E1", "E2", "G12")
) -> SamplingPointFigure:
    """Create a standard polar plot to visualize the polar properties of the laminate."""
    if not sampling_point.is_uptodate or not sampling_point.results:
        raise RuntimeError(f"Sampling point {sampling_point.name} is out-of-date.")

    theta = (
        get_data_from_sp_results(
            "layup", "polar_properties", "angles", results=sampling_point.results
        )
        / 180.0
        * np.pi
    )
    fig, ax = plt.subplots(subplot_kw={"projection": "polar"})

    for comp in components:
        ax.plot(theta, getattr(sampling_point, f"polar_properties_{comp}", []), label=comp)

    ax.set_title("Polar Properties")
    ax.legend()
    return SamplingPointFigure(fig, ax)


def add_ply_sequence_to_plot_to_sp(
    sampling_point: SamplingPoint, axes: Any, core_scale_factor: float = 1.0
) -> None:
    """Add the stacking (ply and text) to an axis or plot."""
    offsets = sampling_point.get_offsets_by_spots(
        spots=[Spot.BOTTOM, Spot.TOP], core_scale_factor=core_scale_factor
    )

    if len(offsets) == 0:
        return

    num_spots = 2
    axes.set_ybound(offsets[0], offsets[-1])
    x_bound = axes.get_xbound()
    width = x_bound[1] - x_bound[0]

    for index, ply in enumerate(sampling_point.analysis_plies):
        angle = float(ply["angle"])
        hatch = "x" if ply["is_core"] else ""

        height = offsets[(index + 1) * num_spots - 1] - offsets[index * num_spots]
        origin = (x_bound[0], offsets[index * num_spots])
        axes.add_patch(Rectangle(xy=origin, width=width, height=height, fill=False, hatch=hatch))
        mat = ply["material"]
        th = float(ply["thickness"])
        text = f"{mat}\nangle={angle:.3}, th={th:.3}"
        axes.annotate(
            text=text,
            xy=(origin[0] + width / 2.0, origin[1] + height / 2.0),
            ha="center",
            va="center",
            fontsize=8,
        )


def add_results_to_plot_to_sp(
    sampling_point: SamplingPoint,
    axes: Any,
    components: Sequence[str],
    spots: Collection[Spot] = (Spot.BOTTOM, Spot.TOP),
    core_scale_factor: float = 1.0,
    title: str = "",
    xlabel: str = "",
) -> None:
    """Add results (strain, stress, or failure values) to an ``Axes`` object."""
    indices = sampling_point.get_indices(spots)
    offsets = sampling_point.get_offsets_by_spots(spots, core_scale_factor)

    for comp in components:
        raw_values = getattr(sampling_point, comp)
        if raw_values is None:
            raise RuntimeError(
                f"Component {comp} is not supported. "
                f"See the description for the 'add_results_to_plot' method."
            )
        values = [raw_values[i] for i in indices]
        axes.plot(values, offsets, label=comp)
    if title:
        axes.set_title(title)
    if xlabel:
        axes.set_xlabel(xlabel)

    axes.legend()
    axes.grid()


def get_result_plots_from_sp(
    sampling_point: SamplingPoint,
    strain_components: Sequence[str] = ("e1", "e2", "e3", "e12", "e13", "e23"),
    stress_components: Sequence[str] = ("s1", "s2", "s3", "s12", "s13", "s23"),
    failure_components: Sequence[FailureMeasureEnum] = (
        FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
        FailureMeasureEnum.RESERVE_FACTOR,
        FailureMeasureEnum.MARGIN_OF_SAFETY,
    ),
    show_failure_modes: bool = False,
    create_laminate_plot: bool = True,
    core_scale_factor: float = 1.0,
    spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
) -> SamplingPointFigure:
    """Generate a figure with a grid of axes (plot) for each selected result entity."""
    num_active_plots = int(create_laminate_plot)
    num_active_plots += 1 if len(strain_components) > 0 else 0
    num_active_plots += 1 if len(stress_components) > 0 else 0
    num_active_plots += 1 if len(failure_components) > 0 else 0

    fig = plt.figure()
    gs = fig.add_gridspec(1, num_active_plots, hspace=0, wspace=0)
    axes = gs.subplots(sharex="col", sharey="row")

    def _get_subplot(axes_obj: Any, current_index: int) -> Any:
        try:
            return axes_obj[current_index]
        except TypeError as exc:
            if current_index > 0:
                raise RuntimeError(f"{type(axes_obj).__name__} plot cannot be indexed.") from exc
            return axes_obj
        except IndexError as exc:
            raise RuntimeError("get_subplot: index exceeds limit.") from exc

    if num_active_plots > 0:
        ticks = sampling_point.get_offsets_by_spots(
            spots=[Spot.TOP], core_scale_factor=core_scale_factor
        )

        axes_index = 0

        first_axis = _get_subplot(axes, axes_index)
        if core_scale_factor != 1.0:
            labels = []
            first_axis.set_ylabel("z-Coordinates (scaled)")
        else:
            labels = [f"{t:.3}" for t in ticks]
            first_axis.set_ylabel("z-Coordinates [length]")

        first_axis.set_yticks(ticks=ticks, labels=labels)

        if create_laminate_plot:
            plt.rcParams["hatch.linewidth"] = 0.2
            plt.rcParams["hatch.color"] = "silver"
            layup_axis = _get_subplot(axes, axes_index)
            sampling_point.add_ply_sequence_to_plot(layup_axis, core_scale_factor)
            layup_axis.set_xticks([])
            axes_index += 1

            plt.rcParams["hatch.linewidth"] = 1.0
            plt.rcParams["hatch.color"] = "black"

        if len(strain_components) > 0:
            strain_axis = _get_subplot(axes, axes_index)
            sampling_point.add_results_to_plot(
                strain_axis, strain_components, spots, core_scale_factor, "Strains", "[-]"
            )
            axes_index += 1

        if len(stress_components) > 0:
            stress_axis = _get_subplot(axes, axes_index)
            sampling_point.add_results_to_plot(
                stress_axis,
                stress_components,
                spots,
                core_scale_factor,
                "Stresses",
                "[force/area]",
            )
            axes_index += 1

        if len(failure_components) > 0:
            failure_axis = _get_subplot(axes, axes_index)
            internal_fc = [FAILURE_MODE_NAMES_TO_ACP[v] for v in failure_components]
            sampling_point.add_results_to_plot(
                failure_axis, internal_fc, spots, core_scale_factor, "Failures", "[-]"
            )

            if show_failure_modes:
                middle_offsets = sampling_point.get_offsets_by_spots(
                    spots=[Spot.MIDDLE], core_scale_factor=core_scale_factor
                )
                critical_failures = sampling_point.get_ply_wise_critical_failures()

                if len(critical_failures) != len(middle_offsets):
                    raise IndexError("Sizes of failures and offsets do not match.")

                for index, offset in enumerate(middle_offsets):
                    for fc in failure_components:
                        failure_axis.annotate(
                            getattr(critical_failures[index], "mode"),
                            xy=(getattr(critical_failures[index], fc.value), offset),
                            xytext=(getattr(critical_failures[index], fc.value), offset),
                        )

            axes_index += 1

    return SamplingPointFigure(fig, axes)



"""Composite data sources."""
from dataclasses import dataclass
import os
import pathlib
from typing import Callable, Dict, List, Optional, Sequence, Union, cast

from ansys.dpf.core import DataSources

from ._typing_helper import PATH as _PATH

__all__ = (
    "CompositeDefinitionFiles",
    "ContinuousFiberCompositesFiles",
    "ShortFiberCompositesFiles",
    "CompositeDataSources",
    "get_composite_files_from_workbench_result_folder",
    "composite_files_from_workbench_harmonic_analysis",
    "get_composites_data_sources",
    "get_short_fiber_composites_data_sources",
)

_SOLID_COMPOSITE_DEFINITIONS_PREFIX = "ACPSolidModel"
_SHELL_COMPOSITE_DEFINITIONS_PREFIX = "ACPCompositeDefinitions"
_SETUP_FOLDER_PREFIX = "Setup"
_H5_SUFFIX = ".h5"
_EXT_SUFFIX = "_ext"
_MATML_FILENAME = "MatML.xml"
_RST_SUFFIX = ".rst"
_RST_PREFIX = "file"
_MAPPING_SUFFIX = ".mapping"

_LAYUPFILE_INDEX_KEY = "CompositeDefinitions"
_MAPPINGFILE_INDEX_KEY = "MappingCompositeDefinitions"


@dataclass
class CompositeDefinitionFiles:
    """Provides the container for composite definition file paths."""

    definition: _PATH
    mapping: Optional[_PATH] = None


@dataclass
class ContinuousFiberCompositesFiles:
    """Provides the container for continuous fiber composite file paths."""

    rst: List[_PATH]
    composite: Dict[str, CompositeDefinitionFiles]
    engineering_data: _PATH
    # True if files are local and false if files
    # have already been uploaded to the server
    files_are_local: bool = True

    def __init__(
        self,
        rst: Union[List[_PATH], _PATH],
        composite: Dict[str, CompositeDefinitionFiles],
        engineering_data: _PATH,
        files_are_local: bool = True,
    ) -> None:
        """Initialize the ContinuousFiberCompositesFiles container.

        Parameters
        ----------
        rst :
            A single path to an RST file, or a list of paths to distributed
            RST files.
        composite :
            Dictionary of composite definition files. The key can be chosen
            freely.
        engineering_data :
            Path to the engineering data file.
        files_are_local :
            True if files are on the local machine, False if they have already
            been uploaded to the DPF server..
        """
        if isinstance(rst, (str, pathlib.Path)):
            rst = [rst]
        self.rst = rst  # type: ignore
        self.composite = composite
        self.engineering_data = engineering_data
        self.files_are_local = files_are_local


@dataclass
class ShortFiberCompositesFiles:
    """Provides the container for short fiber composite file paths."""

    rst: List[_PATH]
    dsdat: _PATH
    engineering_data: _PATH
    # True if files are local and false if files
    # have already been uploaded to the server
    files_are_local: bool = True

    def __init__(
        self,
        rst: Union[List[_PATH], _PATH],
        dsdat: _PATH,
        engineering_data: _PATH,
        files_are_local: bool = True,
    ) -> None:
        """Initialize the ShortFiberCompositesFiles container.

        Parameters
        ----------
        rst :
            A single path to an RST file, or a list of paths to distributed
            RST files.
        dsdat :
            Path to the solver input file (``ds.dat``).
        engineering_data :
            Path to the engineering data file.
        files_are_local :
            True if files are on the local machine, False if they have already
            been uploaded to the DPF server..
        """
        if isinstance(rst, (str, pathlib.Path)):
            rst = [rst]
        self.rst = rst  # type: ignore
        self.dsdat = dsdat
        self.engineering_data = engineering_data
        self.files_are_local = files_are_local


# roosre June 2023: todo add deprecation warning where composite definition label is used
@dataclass(frozen=True)
class CompositeDataSources:
    """Provides data sources related to the composite lay-up.

    Parameters
    ----------
    rst:

    material_support:
        NOTE: The 'material_support' is explicitly listed since it is currently not
        supported (by the DPF Core) to use a distributed RST file as source for the
        material support. Instead, we create a separate DataSources object for the
        material support from the first RST file. This is a workaround until the
        support for distributed RST is added.
    engineering_data:

    old_composite_sources:
        This member is used to support assemblies in combination with the old
        DPF server (<7.0). It should be removed once the support of this
        server version is dropped.

    """

    rst: DataSources
    material_support: DataSources
    composite: DataSources
    engineering_data: DataSources

    old_composite_sources: Dict[str, DataSources]


def _get_mapping_path_file_from_definitions_path_if_exists(
    definition_path: pathlib.Path,
) -> Optional[pathlib.Path]:
    mapping_path = definition_path.parent / (definition_path.stem + _MAPPING_SUFFIX)
    return mapping_path if mapping_path.is_file() else None


def _is_rst_file(path: pathlib.Path) -> bool:
    return path.name.startswith(_RST_PREFIX) and path.suffix == _RST_SUFFIX and path.is_file()


def _is_matml_file(path: pathlib.Path) -> bool:
    return path.name == _MATML_FILENAME and path.is_file()


def _has_ext_suffix(path: pathlib.Path) -> bool:
    """Check if the stem (filename without extension) ends with ``_ext``.

    Example: CompositeDefinitions.1_ext.h5
    """
    return path.stem.endswith("_ext")


def _is_composite_definition_file(path: pathlib.Path) -> bool:
    is_composite_def = path.name.startswith(_SHELL_COMPOSITE_DEFINITIONS_PREFIX)
    no_ext_suffix = not _has_ext_suffix(path)
    return path.suffix == _H5_SUFFIX and path.is_file() and is_composite_def and no_ext_suffix


def _is_solid_model_composite_definition_file(path: pathlib.Path) -> bool:
    is_h5 = path.suffix == _H5_SUFFIX
    has_ext_suffix = _has_ext_suffix(path)
    is_file = path.is_file()
    is_def = path.name.startswith(_SOLID_COMPOSITE_DEFINITIONS_PREFIX)
    return is_h5 and is_file and is_def and not has_ext_suffix


def _get_file_paths_with_predicate(
    predicate: Callable[[pathlib.Path], bool],
    folder: pathlib.Path,
) -> Sequence[pathlib.Path]:
    files = [
        file_or_folder_path
        for file_or_folder_path in folder.iterdir()
        if predicate(file_or_folder_path)
    ]
    return files


def _get_single_file_path_with_predicate(
    predicate: Callable[[pathlib.Path], bool],
    folder: pathlib.Path,
    descriptive_name: str,
) -> pathlib.Path:
    files = _get_file_paths_with_predicate(predicate, folder)
    if len(files) != 1:
        raise RuntimeError(
            f"Expected exactly one {descriptive_name} file. Found {files}."
            f" Available files in folder: {os.listdir(folder)}"
        )

    return files[0]


def _get_single_optional_file_path_with_predicate(
    predicate: Callable[[pathlib.Path], bool],
    folder: pathlib.Path,
    descriptive_name: str,
) -> Optional[pathlib.Path]:
    files = _get_file_paths_with_predicate(predicate, folder)
    if len(files) > 1:
        raise RuntimeError(f"Expected no more than one {descriptive_name} file. Found {files}.")
    if len(files) == 0:
        return None
    return files[0]


def _add_composite_definitons_from_setup_folder(
    setup_folder: pathlib.Path, composite_files: ContinuousFiberCompositesFiles
) -> None:
    composite_definition = _get_single_optional_file_path_with_predicate(
        _is_composite_definition_file,
        setup_folder,
        "composites_definition",
    )

    solid_model_definition = _get_single_optional_file_path_with_predicate(
        _is_solid_model_composite_definition_file, setup_folder, "solid_model_definition"
    )

    def get_composite_definitions_files(
        composite_definition_path: pathlib.Path,
    ) -> CompositeDefinitionFiles:
        mapping_path = _get_mapping_path_file_from_definitions_path_if_exists(
            composite_definition_path
        )
        return CompositeDefinitionFiles(
            definition=composite_definition_path.resolve(),
            mapping=mapping_path.resolve() if mapping_path is not None else None,
        )

    if composite_definition is not None:
        definition_files = get_composite_definitions_files(composite_definition)
        key = os.path.basename(setup_folder) + "_shell"
        if key in composite_files.composite:
            raise RuntimeError(f"Definition with key already exists {key}")
        composite_files.composite[key] = definition_files

    if solid_model_definition is not None:
        definition_files = get_composite_definitions_files(solid_model_definition)
        key = os.path.basename(setup_folder) + "_solid"

        if key in composite_files.composite:
            raise RuntimeError(f"Definition with key already exists {key}")
        composite_files.composite[key] = definition_files


def composite_files_from_workbench_harmonic_analysis(
    result_folder_modal: _PATH, result_folder_harmonic: _PATH
) -> ContinuousFiberCompositesFiles:
    """Get a ``ContinuousFiberCompositesFiles`` object for a harmonic analysis.

    Parameters
    ----------
    result_folder_modal : str
       Result folder of the Modal solution.
       In the Modal system, right-click the **solution** item in the Ansys Mechanical tree
       and select **Open Solver Files Directory** to obtain the result folder.
    result_folder_harmonic : str
       Result folder of the Harmonic Response solution.
       In the Harmonic Response system,
       right-click the **solution** item in the Ansys Mechanical tree
       and select **Open Solver Files Directory** to obtain the result folder.

    """
    result_folder_path_harmonic = pathlib.Path(result_folder_harmonic)
    result_folder_path_modal = pathlib.Path(result_folder_modal)

    setup_folders_modal = [
        folder_path
        for folder_path in result_folder_path_modal.iterdir()
        if folder_path.is_dir() and folder_path.name.startswith(_SETUP_FOLDER_PREFIX)
    ]

    rst_paths = _get_file_paths_with_predicate(
        _is_rst_file,
        result_folder_path_harmonic,
    )

    if len(rst_paths) == 0:
        raise RuntimeError(
            f"Expected at least one rst file. Found {rst_paths}."
            f" Available files in folder: {os.listdir(result_folder_path_harmonic)}"
        )

    matml_path = _get_single_file_path_with_predicate(
        _is_matml_file,
        result_folder_path_harmonic,
        "matml",
    )

    assert matml_path is not None
    assert rst_paths is not None

    continuous_fiber_composite_files = ContinuousFiberCompositesFiles(
        rst=[rst_path.resolve() for rst_path in rst_paths],
        composite={},
        engineering_data=matml_path.resolve(),
    )

    for setup_folder in setup_folders_modal:
        _add_composite_definitons_from_setup_folder(setup_folder, continuous_fiber_composite_files)

    return continuous_fiber_composite_files


def get_composite_files_from_workbench_result_folder(
    result_folder: _PATH, ensure_composite_definitions_found: bool = True
) -> ContinuousFiberCompositesFiles:
    r"""Get a ``ContinuousFiberCompositesFiles`` object from a result folder.

    This function assumes a typical Workbench folder structure for a composite
    simulation. If this method is not able to build the ``ContinuousFiberCompositesFiles``
    object, you can follow these steps:

    In the main Workbench window, activate the files panel by selecting
    **View > Files**. This shows the location of all files used in the
    workbench project. You can determine the different attributes of the
    ``ContinuousFiberCompositesFiles`` object:

    -   ``rst``: A list of files containing either the single ``file.rst``
        file that belongs to the cell ID of the solution, or the distributed
        ``file0.rst`` to ``fileN.rst`` files.

    -   ``engineering_data``: The ``MatML.xml`` file in the same folder as the RST file.

    -   ``composite``: There can be multiple composite definitions,
        one definition for each ACP system if shell data is transferred
        and one definition for each solid model if solid data is transferred.
        All the ``ACPCompositeDefinitions.h5`` and ``ACPSolidModel*.h5``
        files that are used in the solution must be added to the
        ``ContinuousFiberCompositesFiles.composite`` dictionary.
        The key can be chosen freely. Next to the ``ACPCompositeDefinitions.h5``
        and ``ACPSolidModel\*.h5`` files, corresponding ``ACPCompositeDefinitions.mapping``
        and ``ACPSolidModel*.mapping`` files can be found (optional).
        If they exist, they must be added as well.

    The following example shows how a
    :class:`.ContinuousFiberCompositesFiles` object can be built.
    The project in this example has two **ACP Pre** systems, one that exports
    shell information and one that exports solid information.

    The files are located in these locations:

    Result file:

    - ``project_root_folder/dp0/SYS/MECH/file.rst``

    Engineering data file:

    - ``project_root_folder/dp0/SYS/MECH/MatML.xml``

    Composite definition and mapping files for the solid model:

    - ``project_root_folder/dp0/ACP-Pre-1/ACPSolidModel_SM.h5``
    - ``project_root_folder/dp0/ACP-Pre-1/ACPSolidModel_SM.mapping``

    Composite definition and mapping files for the shell model:

    - ``project_root_folder/dp0/ACP-Pre-2/ACPCompositeDefinitions.h5``
    - ``project_root_folder/dp0/ACP-Pre-2/ACPCompositeDefinitions.mapping``

    The code creates the corresponding ``ContinuousFiberCompositesFiles`` object::

        ContinuousFiberCompositesFiles(
            rst=["project_root_folder/dp0/SYS/MECH/file.rst"],
            composite={
               "solid": CompositeDefinitionFiles(
                    definition="project_root_folder/dp0/ACP-Pre-1/ACPSolidModel_SM.h5",
                    mapping="project_root_folder/dp0/ACP-Pre-1/ACPSolidModel_SM.mapping"
                ),
               "shell": CompositeDefinitionFiles(
                    definition="project_root_folder/dp0/ACP-Pre-2/ACPCompositeDefinitions.h5",
                    mapping="project_root_folder/dp0/ACP-Pre-2/ACPCompositeDefinitions.mapping"
                )
            },
            engineering_data="project_root_folder/dp0/SYS/MECH/MatML.xml"
        )

    Parameters
    ----------
    result_folder :
       Result folder of the solution. Right-click the **solution** item in the Ansys Mechanical tree
       and select **Open Solver Files Directory** to obtain the result folder.
    ensure_composite_definitions_found :
        Whether to check if at least one composite definition (shell or solid) has been found.
    """
    result_folder_path = pathlib.Path(result_folder)

    setup_folders = [
        folder_path
        for folder_path in result_folder_path.iterdir()
        if folder_path.is_dir() and folder_path.name.startswith(_SETUP_FOLDER_PREFIX)
    ]

    rst_paths = _get_file_paths_with_predicate(
        _is_rst_file,
        result_folder_path,
    )

    if len(rst_paths) == 0:
        raise RuntimeError(
            f"Expected at least one rst file. Found {rst_paths}."
            f" Available files in folder: {os.listdir(result_folder_path)}"
        )

    matml_path = _get_single_file_path_with_predicate(
        _is_matml_file,
        result_folder_path,
        "matml",
    )

    assert matml_path is not None
    assert rst_paths is not None

    continuous_fiber_composite_files = ContinuousFiberCompositesFiles(
        rst=[rst_path.resolve() for rst_path in rst_paths],
        composite={},
        engineering_data=matml_path.resolve(),
    )

    for setup_folder in setup_folders:
        _add_composite_definitons_from_setup_folder(setup_folder, continuous_fiber_composite_files)

    if (
        ensure_composite_definitions_found
        and len(list(continuous_fiber_composite_files.composite.keys())) == 0
    ):
        raise RuntimeError(
            "No composite definitions found. Set "
            "ensure_composite_definitions_found argument"
            " to False to skip this check. Note: Use the function"
            " composite_files_from_workbench_harmonic_analysis if this is a harmonic analysis."
        )

    return continuous_fiber_composite_files


def _get_data_sources_from_rst_files(
    rst_files: List[_PATH],
) -> DataSources:
    """Get a DPF data sources object from a list of RST files.

    If a single RST file is provided, it is added via 'set_result_file_path'.
    Otherwise, the RST files are added via 'set_domain_result_file_path'.
    """
    data_sources = DataSources()
    if len(rst_files) == 1:
        data_sources.set_result_file_path(rst_files[0])
    else:
        for idx, rst_file in enumerate(rst_files):
            data_sources.set_domain_result_file_path(rst_file, idx)
    return data_sources


def get_composites_data_sources(
    continuous_composite_files: ContinuousFiberCompositesFiles,
) -> CompositeDataSources:
    """Create DPF data sources from a ``ContinuousFiberCompositeFiles`` object.

    Parameters
    ----------
    continuous_composite_files
    """
    if not continuous_composite_files.rst:
        raise RuntimeError("No rst files found.")
    else:
        rst_data_source = _get_data_sources_from_rst_files(continuous_composite_files.rst)

        # NOTE: The 'material_support' is explicitly listed since it is currently not
        # supported (by the DPF Core) to use a distributed RST file as source for the
        # material support. Instead, we create a separate DataSources object for the
        # material support from the first RST file. This is a workaround until the
        # support for distributed RST is added.
        if len(continuous_composite_files.rst) == 1:
            material_support_data_source = rst_data_source
        else:
            material_support_data_source = _get_data_sources_from_rst_files(
                [continuous_composite_files.rst[0]]
            )

    engineering_data_source = DataSources()
    engineering_data_source.add_file_path(
        continuous_composite_files.engineering_data, "EngineeringData"
    )

    composite_data_source = DataSources()
    set_part_key = len(continuous_composite_files.composite) > 1
    old_composite_data_sources = {}

    for key, composite_files in continuous_composite_files.composite.items():
        old_datasource = DataSources()
        if set_part_key:
            composite_data_source.add_file_path_for_specified_result(
                composite_files.definition, _LAYUPFILE_INDEX_KEY, key
            )
        else:
            composite_data_source.add_file_path(composite_files.definition, _LAYUPFILE_INDEX_KEY)

        if composite_files.mapping is not None:
            if set_part_key:
                composite_data_source.add_file_path_for_specified_result(
                    composite_files.mapping, _MAPPINGFILE_INDEX_KEY, key
                )
            else:
                composite_data_source.add_file_path(composite_files.mapping, _MAPPINGFILE_INDEX_KEY)

        ##### This block is needed to support DPF Server older than 7.0 (2023 R2 or before)
        old_datasource = DataSources()
        old_datasource.add_file_path(composite_files.definition, _LAYUPFILE_INDEX_KEY)
        if composite_files.mapping is not None:
            old_datasource.add_file_path(composite_files.mapping, _MAPPINGFILE_INDEX_KEY)
        ##### End of block

        old_composite_data_sources[key] = old_datasource

    return CompositeDataSources(
        rst=rst_data_source,
        material_support=material_support_data_source,
        composite=composite_data_source,
        engineering_data=engineering_data_source,
        old_composite_sources=old_composite_data_sources,
    )


def _data_sources_num_result_keys(data_sources: DataSources) -> int:
    # pylint: disable=protected-access
    return cast(
        int,
        data_sources._api.data_sources_get_num_result_keys(data_sources),
    )


def _data_sources_result_key(data_sources: DataSources, index: int) -> str:
    # pylint: disable=protected-access
    return cast(
        str,
        data_sources._api.data_sources_get_result_key_by_index(data_sources, index),
    )


def get_short_fiber_composites_data_sources(
    short_fiber_composites_files: ShortFiberCompositesFiles,
) -> DataSources:
    """
    Create DPF data sources from a ``ShortFiberCompositeFiles`` object.

    Parameters
    ----------
    short_fiber_composite_files :
        Container for short fiber composite file paths.
    """
    data_sources = _get_data_sources_from_rst_files(short_fiber_composites_files.rst)
    data_sources.add_file_path(short_fiber_composites_files.dsdat, "dat")
    data_sources.add_file_path(short_fiber_composites_files.engineering_data, "EngineeringData")
    return data_sources



"""Helper to get example files."""

from dataclasses import dataclass
import os
import tempfile
from typing import Dict, List, Optional, cast
import urllib.request

import ansys.dpf.core as dpf

from ..data_sources import (
    CompositeDefinitionFiles,
    ContinuousFiberCompositesFiles,
    ShortFiberCompositesFiles,
)

EXAMPLE_REPO = "https://github.com/ansys/example-data/raw/master/pydpf-composites/"


@dataclass
class _ContinuousFiberCompositeFiles:
    definition: str
    mapping: Optional[str] = None


@dataclass
class _ContinuousFiberCompositesExampleFilenames:
    rst: List[str]
    composite: Dict[str, _ContinuousFiberCompositeFiles]
    engineering_data: str


@dataclass
class _ShortFiberCompositesExampleFilenames:
    rst: List[str]
    dsdat: str
    engineering_data: str


@dataclass
class _ContinuousFiberExampleLocation:
    """Location of the a given continuous fiber example in the example_data repo.

    Parameters
    ----------
    directory
        Directory in example_data/pydpf-composites
    files
        Example files in directory
    """

    directory: str
    files: _ContinuousFiberCompositesExampleFilenames


@dataclass
class _ShortFiberExampleLocation:
    """Location of a given short fiber example in the example_data repo.

    Parameters
    ----------
    directory:
        Directory in example_data/pydpf-composites
    files:
        Example files in directory
    """

    directory: str
    files: _ShortFiberCompositesExampleFilenames


_continuous_fiber_examples: Dict[str, _ContinuousFiberExampleLocation] = {
    "shell": _ContinuousFiberExampleLocation(
        directory="shell",
        files=_ContinuousFiberCompositesExampleFilenames(
            rst=["shell.rst"],
            engineering_data="material.engd",
            composite={
                "shell": _ContinuousFiberCompositeFiles(
                    definition="ACPCompositeDefinitions.h5",
                )
            },
        ),
    ),
    "ins": _ContinuousFiberExampleLocation(
        directory="ins",
        files=_ContinuousFiberCompositesExampleFilenames(
            rst=["beam_181_analysis_model.rst"],
            engineering_data="materials.xml",
            composite={
                "shell": _ContinuousFiberCompositeFiles(definition="ACPCompositeDefinitions.h5")
            },
        ),
    ),
    "assembly": _ContinuousFiberExampleLocation(
        directory="assembly",
        files=_ContinuousFiberCompositesExampleFilenames(
            rst=["file.rst"],
            engineering_data="material.engd",
            composite={
                "solid": _ContinuousFiberCompositeFiles(
                    definition="ACPSolidModel_SM.h5", mapping="ACPSolidModel_SM.mapping"
                ),
                "shell": _ContinuousFiberCompositeFiles(
                    definition="ACPCompositeDefinitions.h5",
                    mapping="ACPCompositeDefinitions.mapping",
                ),
            },
        ),
    ),
    "harmonic": _ContinuousFiberExampleLocation(
        directory="harmonic",
        files=_ContinuousFiberCompositesExampleFilenames(
            rst=["file.rst"],
            engineering_data="MatML.xml",
            composite={
                "shell": _ContinuousFiberCompositeFiles(definition="ACPCompositeDefinitions.h5"),
            },
        ),
    ),
}

_short_fiber_examples: Dict[str, _ShortFiberExampleLocation] = {
    "short_fiber": _ShortFiberExampleLocation(
        directory="short_fiber",
        files=_ShortFiberCompositesExampleFilenames(
            rst=["file.rst"], engineering_data="MatML.xml", dsdat="ds.dat"
        ),
    )
}


def _get_file_url(directory: str, filename: str) -> str:
    return EXAMPLE_REPO + "/".join([directory, filename])


def _download_and_upload_file(
    directory: str, filename: str, tmpdir: str, server: dpf.server
) -> str:
    """Download example file from example_data repo and upload it the dpf server."""
    file_url = _get_file_url(directory, filename)
    local_path = os.path.join(tmpdir, filename)
    if server.local_server:
        local_path = os.path.join(os.getcwd(), filename)
    urllib.request.urlretrieve(file_url, local_path)
    if server.local_server:
        return local_path
    return cast(str, dpf.upload_file_in_tmp_folder(local_path, server=server))


def get_short_fiber_example_files(
    server: dpf.server,
    example_key: str,
) -> ShortFiberCompositesFiles:
    """Get short fiber example file by example key.

    This will copy the example files into the current working directory, if the
    server is local.

    Parameters
    ----------
    server:
    example_key:
    """
    example_files = _short_fiber_examples[example_key]
    with tempfile.TemporaryDirectory() as tmpdir:

        def get_server_path(filename: str) -> str:
            return _download_and_upload_file(example_files.directory, filename, tmpdir, server)

        return ShortFiberCompositesFiles(
            rst=[get_server_path(filename) for filename in example_files.files.rst],
            dsdat=get_server_path(example_files.files.dsdat),
            engineering_data=get_server_path(example_files.files.engineering_data),
            files_are_local=False,
        )


def get_continuous_fiber_example_files(
    server: dpf.server,
    example_key: str,
) -> ContinuousFiberCompositesFiles:
    """Get continuous fiber example file by example key.

    This will copy the example files into the current working directory, if the
    server is local.

    Parameters
    ----------
    server:
    example_key:
    """
    example_files = _continuous_fiber_examples[example_key]
    with tempfile.TemporaryDirectory() as tmpdir:

        def get_server_path(filename: str) -> str:
            return _download_and_upload_file(example_files.directory, filename, tmpdir, server)

        all_composite_files = {}
        for key, composite_examples_files_for_scope in example_files.files.composite.items():
            composite_files = CompositeDefinitionFiles(
                definition=get_server_path(composite_examples_files_for_scope.definition),
            )
            if composite_examples_files_for_scope.mapping is not None:
                composite_files.mapping = get_server_path(
                    composite_examples_files_for_scope.mapping
                )

            all_composite_files[key] = composite_files

        return ContinuousFiberCompositesFiles(
            rst=[get_server_path(rst_path) for rst_path in example_files.files.rst],
            engineering_data=get_server_path(example_files.files.engineering_data),
            composite=all_composite_files,
            files_are_local=False,
        )



from enum import Enum

# Use the functional API because otherwise it is impossible to use
# python keywords as enum values (such as "as")
# Note: This enum should to be in sync with the corresponding enum in dpf_composites.
FailureModeEnum = Enum(
    "FailureModeEnum",
    [
        # Strain failure criteria
        ("emax", 101),
        ("evm", 102),
        ("e1", 110),
        ("e1t", 111),
        ("e1c", 112),
        ("e2", 120),
        ("e2t", 121),
        ("e2c", 122),
        ("e3", 130),
        ("e3t", 131),
        ("e3c", 132),
        ("e12", 140),
        ("e13", 150),
        ("e23", 160),
        # Stress failure criteria
        ("smax", 201),
        ("svm", 202),
        ("s1", 210),
        ("s1t", 211),
        ("s1c", 212),
        ("s2", 220),
        ("s2t", 221),
        ("s2c", 222),
        ("s3", 230),
        ("s3t", 231),
        ("s3c", 232),
        ("s12", 240),
        ("s13", 250),
        ("s23", 260),
        # Sandwich failure criteria
        ("cf", 310),  # core shear
        ("w", 320),  # wrinkling
        ("wb", 321),  # wrinkling bot face sheet
        ("wt", 322),  # wrinkling top face sheet
        ("sc", 330),  # shear crimping
        # Classical failure criteria
        # from laminate theories
        ("tw", 400),  # tsai wu
        ("th", 500),  # tsai hill
        ("h", 501),  # hill
        ("hf", 601),  # hashin fiber
        ("hm", 602),  # hashin matrix
        ("hd", 603),  # hashin delamination
        ("ho", 700),  # hoffmann
        ("p", 800),  # puck
        ("pf", 801),  # puck fiber
        ("pmA", 802),  # puck matrix modus A
        ("pmB", 803),  # puck matrix modus B
        ("pmC", 804),  # puck matrix modus C
        ("pd", 805),  # puck 3D delamination
        ("l", 900),  # larc failure criterion
        ("lft3", 901),  # larc fiber failure tension LaRC  # 3
        ("lfc4", 902),  # larc fiber failure compression LaRC  # 4
        ("lfc6", 903),  # larc fiber failure compression LaRC  # 6
        ("lmt1", 904),  # larc matrix tension failure LaRC  # 1
        ("lmc2", 905),  # larc matrix compression failure LaRC  # 2
        ("lmc5", 906),  # larc matrix compression failure LaRC  # 5
        ("c", 1000),  # cuntze
        ("cft", 1001),  # cuntze fiber failure tension
        ("cfc", 1002),  # cuntze fiber failure compression
        ("cmA", 1003),  # cuntze matrix failure tension
        ("cmB", 1004),  # cuntze matrix failure compression
        ("cmC", 1005),  # cuntze wedge shape matrix failure
        ("vMe", 1101),  # elastic von mises strain failure
        ("vMs", 1102),  # elastic von mises stress failure
        ("as", 1201),  # adhesive shear stress failure
        ("ap", 1202),  # adhesive peel stress failure
        ("af", 1203),  # adhesive fracture stress failure
        # Allow to  mark text as ns = no show in order to not show
        # failure mode labels below / above a threshold etc.
        ("ns", 9998),
        # If no failure mode can be computed, explicitly assign n/a - not available.
        ("na", 9999),
    ],
)



"""Unit system helper."""
__all__ = ("get_unit_system", "UnitSystemProvider")

from typing import Optional, Union

from ansys.dpf.core import DataSources, Operator, UnitSystem

# Either a UnitSystem or a ResultInfoProvider operator.
# This Union is needed because dpf currently does not support passing
# a ResultInfo directly to operators nor is it possible to get
# the UnitSystem from a ResultInfo.
UnitSystemProvider = Union[UnitSystem, Operator]


def get_unit_system(
    data_source_or_streams_provider: Union[DataSources, Operator],
    default_unit_system: Optional[UnitSystem] = None,
) -> UnitSystemProvider:
    """Get unit_system from rst DataSources.

    Returns the ResultInfo from the result file or
    the default_unit_system if no unit system is found in the result file
    (for example for pure mapdl result files).

    Parameters
    ----------
    data_source_or_streams_provider:
        DPF Data Source or streams provider containing a rst file.
    default_unit_system:
        Default Unit system that is used if the rst file does not contain
        a unit system.
    """
    result_info_provider = Operator("ResultInfoProvider")
    if isinstance(data_source_or_streams_provider, Operator):
        result_info_provider.inputs.streams_container(data_source_or_streams_provider)
    else:
        result_info_provider.inputs.data_sources(data_source_or_streams_provider)

    # If unit system is undefined, fall back to default_unit_system
    if result_info_provider.outputs.result_info().unit_system == "Undefined":
        if default_unit_system is None:
            raise RuntimeError(
                "The result file does not specify a unit system."
                " Please define a default unit system. "
                "A default unit system can be passed when "
                "creating a CompositeModel."
            )
        return default_unit_system
    else:
        return result_info_provider



"""Defines the base class of composite failure criterion."""

import json
from typing import Any, Dict, Sequence


class FailureCriterionBase:
    """Defines the base class for failure criteria."""

    def __init__(self, *, name: str, active: bool):
        """Do not use directly! Construct a base failure criterion."""
        self.active = active
        self._name = name

    def _get_active(self) -> bool:
        return self._active

    def _set_active(self, value: bool) -> None:
        self._active = value

    def _get_name(self) -> str:
        return self._name

    active = property(
        _get_active, _set_active, doc="The failure criterion is suppressed if active is False."
    )
    name = property(_get_name, doc="Name of the failure criterion. Read only.")

    def to_dict(self) -> Dict[str, Any]:
        """:return: a dict with all properties."""
        properties = self._get_properties(exclude=["name"])

        attr_dict = {}
        for prop in properties:
            attr_dict[prop] = getattr(self, prop)

        return attr_dict

    def to_json(self) -> str:
        """:return: the string representation of the object as JSON.

        It can be used for the result definition of the DPF Composites Failure Operator.
        """
        return json.dumps(self.to_dict())

    def _get_properties(self, exclude: Sequence[str] = tuple()) -> Sequence[Any]:
        properties = [
            attr
            for attr in dir(self)
            if not attr.startswith("__")
            and not attr.startswith("_")
            and not callable(getattr(self, attr))
            and attr not in exclude
        ]

        return properties

    def _short_descr(self) -> str:
        """:return: short description of the object."""
        return f"{self.__class__.__name__}(name='{self.name}', active={self.active})"

    def __repr__(self) -> str:
        """:return: string conversion."""
        s_attrs = ", ".join([f"{attr}={getattr(self, attr)}" for attr in self._get_properties()])
        s = f"{self.__class__.__name__}({s_attrs})"
        return s



"""Composite Scope."""
from dataclasses import dataclass
from typing import Optional, Sequence


@dataclass(frozen=True)
class CompositeScope:
    """Provides the composite scope.

    This class defines which part of the model and solution step are selected.

    Parameters
    ----------
    elements:
        List of elements.
    plies:
        List of plies.
    time:
        Time or frequency. You can use the
        :meth:`.CompositeModel.get_result_times_or_frequencies` method
        to list the solution steps.
    named_selections:
        List of element sets.
        Use `composite_model.get_mesh().available_named_selections` to list
        all named selections.

    Notes
    -----
    If more than one scope (``elements``, ``named_selections`` and ``plies``)
    is set, then the final element scope is the intersection
    of the defined parameters. All elements are selected if no parameter is set.

    """

    elements: Optional[Sequence[int]] = None
    plies: Optional[Sequence[str]] = None
    time: Optional[float] = None
    named_selections: Optional[Sequence[str]] = None



"""Defines the MaxStrain failure criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_E1 = "Activates the failure evaluation regarding the strain in the material 1 direction."
_DOC_E2 = "Activates the failure evaluation regarding the strain in the material 2 direction."
_DOC_E3 = (
    "Activates the failure evaluation regarding the strain in the "
    "material 3 direction (out-of-plane)."
)
_DOC_E12 = "Whether to activate the failure evaluation regarding the in-plane shear strain e12."
_DOC_E13 = "Whether to activate the failure evaluation regarding the interlaminar shear strain e13."
_DOC_E23 = "Whether to activate the failure evaluation regarding the interlaminar shear strain e23."
_DOC_WF_E1 = "Weighting factor of the failure mode e1."
_DOC_WF_E2 = "Weighting factor of the failure mode e2."
_DOC_WF_E3 = "Weighting factor of the failure mode e3."
_DOC_WF_E12 = "Weighting factor of the failure mode e12."
_DOC_WF_E13 = "Weighting factor of the failure mode e13."
_DOC_WF_E23 = "Weighting factor of the failure mode e23."
_DOC_FORCE_GLOBAL_STRAIN_LIMITS = (
    "Whether to use one set of global strain limits instead of the "
    "strain limits of the materials."
)
_DOC_EXT = "Global tensile strain limit in material direction 1."
_DOC_EXC = "Global compressive strain limit in material direction 1."
_DOC_EYT = "Global tensile strain limit in material direction 2."
_DOC_EYC = "Global compressive strain limit in material direction 2."
_DOC_EZT = "Global strain limit in material direction 3."
_DOC_EZC = "Global compressive strain limit in material direction 3."
_DOC_ESXY = "Global strain limit in material direction 12."
_DOC_ESXZ = "Global strain limit in material direction 13."
_DOC_ESYZ = "Global strain limit in material direction 23."


class MaxStrainCriterion(FailureCriterionBase):
    """Max Strain Criterion."""

    __doc__ = f"""Defines the maximum strain failure criterion for orthotropic reinforced materials.

    Parameters
    ----------
    e1:
        {_DOC_E1}
    e2:
        {_DOC_E2}
    e3:
        {_DOC_E3}
    e12:
        {_DOC_E12}
    e13:
        {_DOC_E13}
    e23:
        {_DOC_E23}
    wf_e1:
        {_DOC_WF_E1}
    wf_e2:
        {_DOC_WF_E2}
    wf_e3:
        {_DOC_WF_E3}
    wf_e12:
        {_DOC_WF_E12}
    wf_e13:
        {_DOC_WF_E13}
    wf_e23:
        {_DOC_WF_E23}
    eXt:
        {_DOC_EXT}
    eXc:
        {_DOC_EXC}
    eYt:
        {_DOC_EYT}
    eYc:
        {_DOC_EYC}
    eZt:
        {_DOC_EZT}
    eZc:
        {_DOC_EZC}
    eSxy:
        {_DOC_ESXY}
    eSxz:
        {_DOC_ESXZ}
    eSyz:
        {_DOC_ESYZ}
    """

    def __init__(
        self,
        *,
        e1: bool = True,
        e2: bool = True,
        e3: bool = False,
        e12: bool = True,
        e13: bool = False,
        e23: bool = False,
        wf_e1: float = 1.0,
        wf_e2: float = 1.0,
        wf_e3: float = 1.0,
        wf_e12: float = 1.0,
        wf_e13: float = 1.0,
        wf_e23: float = 1.0,
        force_global_strain_limits: bool = False,
        eXt: float = 0.0,
        eXc: float = 0.0,
        eYt: float = 0.0,
        eYc: float = 0.0,
        eZt: float = 0.0,
        eZc: float = 0.0,
        eSxy: float = 0.0,
        eSxz: float = 0.0,
        eSyz: float = 0.0,
    ):
        """Create a maximum strain failure criterion for orthotropic reinforced materials."""
        super().__init__(name="Max Strain", active=True)

        for attr in ATTRS_MAX_STRAIN:
            setattr(self, attr, eval(attr))

    def _get_e1(self) -> bool:
        return self._e1

    def _set_e1(self, value: bool) -> None:
        self._e1 = value

    def _get_e2(self) -> bool:
        return self._e2

    def _set_e2(self, value: bool) -> None:
        self._e2 = value

    def _get_e3(self) -> bool:
        return self._e3

    def _set_e3(self, value: bool) -> None:
        self._e3 = value

    def _get_e12(self) -> bool:
        return self._e12

    def _set_e12(self, value: bool) -> None:
        self._e12 = value

    def _get_e13(self) -> bool:
        return self._e13

    def _set_e13(self, value: bool) -> None:
        self._e13 = value

    def _get_e23(self) -> bool:
        return self._e23

    def _set_e23(self, value: bool) -> None:
        self._e23 = value

    def _get_wf_e1(self) -> float:
        return self._wf_e1

    def _set_wf_e1(self, value: float) -> None:
        self._wf_e1 = value

    def _get_wf_e2(self) -> float:
        return self._wf_e2

    def _set_wf_e2(self, value: float) -> None:
        self._wf_e2 = value

    def _get_wf_e3(self) -> float:
        return self._wf_e3

    def _set_wf_e3(self, value: float) -> None:
        self._wf_e3 = value

    def _get_wf_e12(self) -> float:
        return self._wf_e12

    def _set_wf_e12(self, value: float) -> None:
        self._wf_e12 = value

    def _get_wf_e13(self) -> float:
        return self._wf_e13

    def _set_wf_e13(self, value: float) -> None:
        self._wf_e13 = value

    def _get_wf_e23(self) -> float:
        return self._wf_e23

    def _set_wf_e23(self, value: float) -> None:
        self._wf_e23 = value

    def _get_force_global_strain_limits(self) -> bool:
        return self._force_global_strain_limits

    def _set_force_global_strain_limits(self, value: bool) -> None:
        self._force_global_strain_limits = value

    def _get_eXt(self) -> float:
        return self._eXt

    def _set_eXt(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Tensile limit eXt cannot be negative.")
        self._eXt = value

    def _get_eXc(self) -> float:
        return self._eXc

    def _set_eXc(self, value: float) -> None:
        if value > 0.0:
            raise ValueError("Compressive limit eXc cannot be positive.")
        self._eXc = value

    def _get_eYt(self) -> float:
        return self._eYt

    def _set_eYt(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Tensile limit eYt cannot be negative.")
        self._eYt = value

    def _get_eYc(self) -> float:
        return self._eYc

    def _set_eYc(self, value: float) -> None:
        if value > 0.0:
            raise ValueError("Compressive limit eYc cannot be positive.")
        self._eYc = value

    def _get_eZt(self) -> float:
        return self._eZt

    def _set_eZt(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Tensile limit eZt cannot be negative.")
        self._eZt = value

    def _get_eZc(self) -> float:
        return self._eZc

    def _set_eZc(self, value: float) -> None:
        if value > 0.0:
            raise ValueError("Compressive limit eZc cannot be positive.")
        self._eZc = value

    def _get_eSxy(self) -> float:
        return self._eSxy

    def _set_eSxy(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Strain limit eSxy cannot be negative.")
        self._eSxy = value

    def _get_eSxz(self) -> float:
        return self._eSxz

    def _set_eSxz(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Strain limit eSxz cannot be negative.")
        self._eSxz = value

    def _get_eSyz(self) -> float:
        return self._eSyz

    def _set_eSyz(self, value: float) -> None:
        if value < 0.0:
            raise ValueError("Strain limit eSyz cannot be negative.")
        self._eSyz = value

    e1 = property(_get_e1, _set_e1, doc=_DOC_E1)
    e2 = property(_get_e2, _set_e2, doc=_DOC_E2)

    e3 = property(_get_e3, _set_e3, doc=_DOC_E3)
    e12 = property(_get_e12, _set_e12, doc=_DOC_E12)
    e13 = property(_get_e13, _set_e13, doc=_DOC_E13)
    e23 = property(_get_e23, _set_e23, doc=_DOC_E23)

    wf_e1 = property(_get_wf_e1, _set_wf_e1, doc=_DOC_WF_E1)
    wf_e2 = property(_get_wf_e2, _set_wf_e2, doc=_DOC_WF_E2)
    wf_e3 = property(_get_wf_e3, _set_wf_e3, doc=_DOC_WF_E3)
    wf_e12 = property(_get_wf_e12, _set_wf_e12, doc=_DOC_WF_E12)
    wf_e13 = property(_get_wf_e13, _set_wf_e13, doc=_DOC_WF_E13)
    wf_e23 = property(_get_wf_e23, _set_wf_e23, doc=_DOC_WF_E23)

    force_global_strain_limits = property(
        _get_force_global_strain_limits,
        _set_force_global_strain_limits,
        doc=_DOC_FORCE_GLOBAL_STRAIN_LIMITS,
    )
    eXt = property(_get_eXt, _set_eXt, doc=_DOC_EXT)
    eXc = property(_get_eXc, _set_eXc, doc=_DOC_EXC)
    eYt = property(_get_eYt, _set_eYt, doc=_DOC_EYT)
    eYc = property(_get_eYc, _set_eYc, doc=_DOC_EXC)
    eZt = property(_get_eZt, _set_eZt, doc=_DOC_EZT)
    eZc = property(_get_eZc, _set_eZc, doc=_DOC_EZC)
    eSxy = property(_get_eSxy, _set_eSxy, doc=_DOC_ESXY)

    eSxz = property(_get_eSxz, _set_eSxz, doc=_DOC_ESXZ)
    eSyz = property(_get_eSyz, _set_eSyz, doc=_DOC_ESYZ)


ATTRS_MAX_STRAIN = inspect.signature(MaxStrainCriterion).parameters.keys()



import dataclasses
from typing import Dict

from ansys.dpf.core.server_types import BaseServer
from packaging import version


@dataclasses.dataclass(frozen=True)
class _DpfVersionInfo:
    server_version: str
    wb_version: str
    description: str


_DPF_VERSIONS: Dict[str, _DpfVersionInfo] = {
    "5.0": _DpfVersionInfo("5.0", "2023 R1", "Initial release of DPF Composites."),
    "7.0": _DpfVersionInfo("7.0", "2024 R1", "DPF Composites plugin with sub-operators."),
}


def _check_key(ver: str) -> None:
    if ver not in _DPF_VERSIONS.keys():
        msg = ", ".join([f"{index}:{ver.server_version}" for index, ver in _DPF_VERSIONS.items()])
        raise RuntimeError(f"Invalid key `{ver}`. Allowed values are {msg}.")


def version_older_than(server: BaseServer, ver: str) -> bool:
    """Evaluate if the dpf server is older than the given version."""
    _check_key(ver)
    version_info = _DPF_VERSIONS[ver]
    return version.parse(server.version) < version.parse(version_info.server_version)


def version_equal_or_later(server: BaseServer, ver: str) -> bool:
    """Evaluate if the dpf server is equal or newer than the given version."""
    _check_key(ver)
    version_info = _DPF_VERSIONS[ver]
    return version.parse(server.version) >= version.parse(version_info.server_version)



from enum import IntEnum


class LayupProperty(IntEnum):
    """Enum for lay-up properties.

    Values correspond to labels in the output container of the lay-up provider.
    """

    ANGLE = 0
    SHEAR_ANGLE = 1
    THICKNESS = 2
    LAMINATE_OFFSET = 3


class LayerProperty(IntEnum):
    """Provides the layer properties available."""

    THICKNESSES = 0
    ANGLES = 1
    SHEAR_ANGLES = 2



"""Utilities for managing the DPF server.

Helper functions for managing the DPF server, in particular for loading
the DPF Composites plugin.
"""

from ._connect_to_or_start_server import connect_to_or_start_server
from ._load_plugin import load_composites_plugin
from ._upload_files_to_server import (
    upload_continuous_fiber_composite_files_to_server,
    upload_short_fiber_composite_files_to_server,
)
from ._versions import version_equal_or_later, version_older_than

__all__ = (
    "load_composites_plugin",
    "connect_to_or_start_server",
    "upload_short_fiber_composite_files_to_server",
    "upload_continuous_fiber_composite_files_to_server",
    "version_older_than",
    "version_equal_or_later",
)



"""Defines the MaxStress failure criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_S1 = (
    "Whether to activate the failure evaluation regarding the stress in the material 1 direction."
)
_DOC_S2 = (
    "Whether to activate the failure evaluation regarding the stress in the material 2 direction."
)
_DOC_S3 = (
    "Whether to activate the failure evaluation regarding the stress in the material "
    "3 direction (out-of-plane)."
)
_DOC_S12 = "Whether to activate the failure evaluation regarding the in-plane shear stress s12."
_DOC_S13 = "Whether to activate the failure evaluation regarding the interlaminar shear stress s13."
_DOC_S23 = "Whether to activate the failure evaluation regarding the interlaminar shear stress s23."
_DOC_WF_S1 = "Weighting factor of the failure mode s1."
_DOC_WF_S2 = "Weighting factor of the failure mode s2."
_DOC_WF_S3 = "Weighting factor of the failure mode s3."
_DOC_WF_S12 = "Weighting factor of the failure mode s12."
_DOC_WF_S13 = "Weighting factor of the failure mode s13."
_DOC_WF_S23 = "Weighting factor of the failure mode s23."


class MaxStressCriterion(FailureCriterionBase):
    """Max Stress Criterion."""

    __doc__ = f"""Defines the maximum stress failure criterion for orthotropic reinforced materials.

    Parameters
    ----------
    s1:
        {_DOC_S1}
    s2:
        {_DOC_S2}
    s3:
        {_DOC_S3}
    s12:
        {_DOC_S12}
    s13:
        {_DOC_S13}
    s23:
        {_DOC_S23}
    wf_s1:
        {_DOC_WF_S1}
    wf_s2:
        {_DOC_WF_S2}
    wf_s3:
        {_DOC_WF_S3}
    wf_s12:
        {_DOC_WF_S12}
    wf_s13:
        {_DOC_WF_S13}
    wf_s23:
        {_DOC_WF_S23}
    """

    def __init__(
        self,
        *,
        s1: bool = True,
        s2: bool = True,
        s3: bool = False,
        s12: bool = True,
        s13: bool = False,
        s23: bool = False,
        wf_s1: float = 1.0,
        wf_s2: float = 1.0,
        wf_s3: float = 1.0,
        wf_s12: float = 1.0,
        wf_s13: float = 1.0,
        wf_s23: float = 1.0,
    ):
        """Create a maximum stress failure criterion for orthotropic reinforced materials."""
        super().__init__(name="Max Stress", active=True)

        for attr in ATTRS_MAX_STRESS:
            setattr(self, attr, eval(attr))

    def _get_s1(self) -> bool:
        return self._s1

    def _set_s1(self, value: bool) -> None:
        self._s1 = value

    def _get_s2(self) -> bool:
        return self._s2

    def _set_s2(self, value: bool) -> None:
        self._s2 = value

    def _get_s3(self) -> bool:
        return self._s3

    def _set_s3(self, value: bool) -> None:
        self._s3 = value

    def _get_s12(self) -> bool:
        return self._s12

    def _set_s12(self, value: bool) -> None:
        self._s12 = value

    def _get_s13(self) -> bool:
        return self._s13

    def _set_s13(self, value: bool) -> None:
        self._s13 = value

    def _get_s23(self) -> bool:
        return self._s23

    def _set_s23(self, value: bool) -> None:
        self._s23 = value

    def _get_wf_s1(self) -> float:
        return self._wf_s1

    def _set_wf_s1(self, value: float) -> None:
        self._wf_s1 = value

    def _get_wf_s2(self) -> float:
        return self._wf_s2

    def _set_wf_s2(self, value: float) -> None:
        self._wf_s2 = value

    def _get_wf_s3(self) -> float:
        return self._wf_s3

    def _set_wf_s3(self, value: float) -> None:
        self._wf_s3 = value

    def _get_wf_s12(self) -> float:
        return self._wf_s12

    def _set_wf_s12(self, value: float) -> None:
        self._wf_s12 = value

    def _get_wf_s13(self) -> float:
        return self._wf_s13

    def _set_wf_s13(self, value: float) -> None:
        self._wf_s13 = value

    def _get_wf_s23(self) -> float:
        return self._wf_s23

    def _set_wf_s23(self, value: float) -> None:
        self._wf_s23 = value

    s1 = property(_get_s1, _set_s1, doc=_DOC_S1)

    s2 = property(_get_s2, _set_s2, doc=_DOC_S2)

    s3 = property(_get_s3, _set_s3, doc=_DOC_S3)
    s12 = property(_get_s12, _set_s12, doc=_DOC_S12)
    s13 = property(_get_s13, _set_s13, doc=_DOC_S13)
    s23 = property(_get_s23, _set_s23, doc=_DOC_S23)

    wf_s1 = property(_get_wf_s1, _set_wf_s1, doc=_DOC_WF_S1)
    wf_s2 = property(_get_wf_s2, _set_wf_s2, doc=_DOC_WF_S2)
    wf_s3 = property(_get_wf_s3, _set_wf_s3, doc=_DOC_WF_S3)
    wf_s12 = property(_get_wf_s12, _set_wf_s12, doc=_DOC_WF_S12)

    wf_s13 = property(_get_wf_s13, _set_wf_s13, doc=_DOC_WF_S13)
    wf_s23 = property(_get_wf_s23, _set_wf_s23, doc=_DOC_WF_S23)


ATTRS_MAX_STRESS = inspect.signature(MaxStressCriterion).parameters.keys()



from typing import cast

import ansys.dpf.core as dpf
from ansys.dpf.core.server_types import BaseServer

from .._typing_helper import PATH as _PATH
from ..data_sources import (
    CompositeDefinitionFiles,
    ContinuousFiberCompositesFiles,
    ShortFiberCompositesFiles,
)


def upload_short_fiber_composite_files_to_server(
    data_files: ShortFiberCompositesFiles, server: BaseServer
) -> ShortFiberCompositesFiles:
    """Upload short fiber composites files to server.

    Parameters
    ----------
    data_files
    server
    """
    # If files are not local, it means they have already been
    # uploaded to the server
    if server.local_server or not data_files.files_are_local:
        return data_files

    def upload(filename: _PATH) -> str:
        return cast(str, dpf.upload_file_in_tmp_folder(filename, server=server))

    return ShortFiberCompositesFiles(
        rst=[upload(filename) for filename in data_files.rst],
        dsdat=upload(data_files.dsdat),
        engineering_data=upload(data_files.engineering_data),
        files_are_local=False,
    )


def upload_continuous_fiber_composite_files_to_server(
    data_files: ContinuousFiberCompositesFiles, server: BaseServer
) -> ContinuousFiberCompositesFiles:
    """Upload continuous fiber composites files to server.

    Note: If server.local_server == True the data_files are returned unmodified.

    Parameters
    ----------
    data_files
    server
    """
    # If files are not local, it means they have already been
    # uploaded to the server
    if server.local_server or not data_files.files_are_local:
        return data_files

    def upload(filename: _PATH) -> _PATH:
        return cast(str, dpf.upload_file_in_tmp_folder(filename, server=server))

    all_composite_files = {}
    for key, composite_files_by_scope in data_files.composite.items():
        composite_definition_files = CompositeDefinitionFiles(
            definition=upload(composite_files_by_scope.definition),
        )

        if composite_files_by_scope.mapping is not None:
            composite_definition_files.mapping = upload(composite_files_by_scope.mapping)

        all_composite_files[key] = composite_definition_files

    return ContinuousFiberCompositesFiles(
        rst=[upload(filename) for filename in data_files.rst],
        engineering_data=upload(data_files.engineering_data),
        composite=all_composite_files,
        files_are_local=False,
    )



"""Helper function to get material-related operators."""
from typing import Optional
from warnings import warn

from ansys.dpf.core import DataSources, Operator

__all__ = ("MaterialOperators", "get_material_operators")

from ansys.dpf.composites.unit_system import UnitSystemProvider


class MaterialOperators:
    """Provides the container for material-related operators.

    Parameters
    ----------
    material_support_provider:
        The material support provider takes care of mapping the materials in the RST file to
        the materials in the composite definitions.
        The material support contains all the materials from the RST file. Currently
        the output of this operator cannot be inspected in Python.
    material_provider:
        Outputs the ``MaterialsContainer``, which can be used to
        evaluate material properties. This container cannot be
        queried in Python, but it can be passed to other DPF operators
        that evaluate the properties.
    result_info_provider:
        Provides the ``ResultInfo`` object.
    """

    def __init__(
        self,
        material_provider: Operator,
        material_support_provider: Operator,
        result_info_provider: Operator,
    ):
        """Create a MaterialOperators object."""
        self._material_provider = material_provider
        self._material_support_provider = material_support_provider
        self._result_info_provider = result_info_provider

    @property
    def result_info_provider(self) -> Operator:
        """Get result_info_provider."""
        warn(
            "Getting the result_info_provider from a "
            "MaterialsOperators object is deprecated. "
            "Use get_unit_system to obtain the unit system.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._result_info_provider

    @property
    def material_support_provider(self) -> Operator:
        """Get material_support_provider."""
        return self._material_support_provider

    @property
    def material_provider(self) -> Operator:
        """Get material_provider."""
        return self._material_provider


def get_material_operators(
    rst_data_source: DataSources,
    engineering_data_source: DataSources,
    unit_system: Optional[UnitSystemProvider] = None,
) -> MaterialOperators:
    """Get material properties related to operators.

    Parameters
    ----------
    rst_data_source
        Data source that contains a RST file. Note that multiple (distributed)
        RST files are not supported.
    engineering_data_source
        Data source that contains the Engineering Data file.
    unit_system:
        Unit System
    ----------

    """
    material_support_provider = Operator("support_provider")
    material_support_provider.inputs.property("mat")
    material_support_provider.inputs.data_sources(rst_data_source)

    result_info_provider = Operator("ResultInfoProvider")
    result_info_provider.inputs.data_sources(rst_data_source)

    if unit_system is None:
        warn(
            "Calling get_material_operators without "
            "a unit_system argument is deprecated. Use get_unit_system "
            "to obtain the unit_system",
            DeprecationWarning,
            stacklevel=2,
        )
        unit_system = result_info_provider.outputs.result_info

    # Set up material provider
    # Combines the material support in the engineering data XML file and the unit system.
    # Its output can be used to evaluate material properties.
    material_provider = Operator("eng_data::ans_mat_material_provider")
    material_provider.inputs.data_sources = engineering_data_source
    material_provider.inputs.unit_system_or_result_info(unit_system)
    material_provider.inputs.abstract_field_support(
        material_support_provider.outputs.abstract_field_support
    )
    material_provider.inputs.Engineering_data_file(engineering_data_source)

    return MaterialOperators(
        material_provider=material_provider,
        material_support_provider=material_support_provider,
        result_info_provider=result_info_provider,
    )



"""Combined Failure Criterion."""

import json
from typing import Any, Dict, Sequence

from ._failure_criterion_base import FailureCriterionBase


class CombinedFailureCriterion:
    """Defines the combined failure criterion.

    This class can be used in combination with the failure evaluator operator in
    DPF Composites.

    Examples
    --------
        >>> combined_failure = CombinedFailureCriterion("max_stress 3D")
        >>> max_stress = MaxStressCriterion(s1=True, s2=True, s3=True, s12=True, s13=True, s23=True)
        >>> combined_failure.insert(max_stress)

    """

    JSON_DICT_KEY = "criteria"

    def __init__(
        self,
        name: str = "CombinedFailureCriterion",
        failure_criteria: Sequence[FailureCriterionBase] = (),
    ):
        """Create a combined failure criterion.

        Parameters
        ----------
        name:
            User-defined name of the criterion.
        failure_criteria:
            List of failure criteria.
        """
        self._failure_criteria: Dict[str, FailureCriterionBase] = {}
        for fc in failure_criteria:
            self.insert(fc)

        self.name = name

    def _get_name(self) -> str:
        return self._name

    def _set_name(self, value: str) -> None:
        self._name = value

    def _get_failure_criteria(self) -> Dict[str, FailureCriterionBase]:
        return self._failure_criteria

    name = property(_get_name, _set_name, doc="Name of the combined failure criterion.")
    failure_criteria = property(
        _get_failure_criteria,
        doc="List of failure criteria. Use insert and remove to edit the list.",
    )

    def insert(self, fc: FailureCriterionBase) -> None:
        """Add a failure criterion to a list of selected criteria.

        Parameters
        ----------
        fc:
            Failure criterion to add. If a failure criterion of the same type
            already exists, it is overwritten.

        Examples
        --------
            >>> combined_failure = CombinedFailureCriterion("max_stress 3D")
            >>> max_stress = MaxStressCriterion(s1=True, s2=True, s3=True,
                                                s12=True, s13=True, s23=True)
            >>> combined_failure.insert(max_stress)

        """
        if fc is not None:
            self._failure_criteria[fc.name] = fc

    def remove(self, key: str) -> FailureCriterionBase:
        """Remove a failure criterion.

        Parameters
        ----------
        key:
            Name of the failure criterion.

        Returns
        -------
        :
            Removed failure criterion or ``None``.

        Examples
        --------
            >>> combined_failure.remove("Max Stress")

        """
        if not key in self._failure_criteria.keys():
            raise KeyError(f"{key} does not exist in the list of failure criteria.")

        return self._failure_criteria.pop(key)

    def to_dict(self) -> Dict[str, Any]:
        """Return the combined failure criterion as a dictionary.

        Returns
        -------
        :
            JSON dictionary that can be used for the result definition
            of the DPF Composites Failure evaluator operator.
        """
        criteria = {}
        for _, fc in self.failure_criteria.items():
            # returns a dict of all attributes
            attr_dict = fc.to_dict()
            # get the name and use it as key to add the attrs
            key = fc.name.lower().replace(" ", "_")
            failure_dict = {key: attr_dict}
            criteria.update(failure_dict)

        return criteria

    def to_json(self) -> str:
        """Return the combined failure criterion as a JSON dictionary.

        Returns
        -------
        :
           String representation (``json.dumps`` file) that can be used for the result definition
           of the DPF Composites Failure evaluator operator.
        """
        return json.dumps(self.to_dict())

    def __repr__(self) -> str:
        """Return a description of the combined failure criteria."""
        s_criteria = ", ".join(
            [f"'{k}': {fc._short_descr()}" for k, fc in self.failure_criteria.items()]
        )
        s = f"{self.__class__.__name__}(failure_criteria={{{s_criteria}}})"
        return s



"""Tsai-Hill failure criterion."""

from ._quadratic_failure_criterion import _DOC_DIM, _DOC_WF, QuadraticFailureCriterion


class TsaiHillCriterion(QuadraticFailureCriterion):
    """Tsai Hill Criterion."""

    __doc__ = f"""Defines the Tsai-Hill failure criterion for orthotropic reinforced materials.

    Parameters
    ----------
    wf:
        {_DOC_WF}
    dim:
        {_DOC_DIM}
    """

    def __init__(self, *, active: bool = True, wf: float = 1.0, dim: int = 2):
        """Create a Tsai-Hill failure criterion for orthotropic reinforced materials."""
        super().__init__(name="Tsai Hill", active=active, dim=dim, wf=wf)



"""Core failure criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_INCLUDE_INS = (
    "Whether to activate the formulation that considers interlaminar normal stresses."
)

_DOC_WF = "Weighting factor of the failure mode (cs)."


class CoreFailureCriterion(FailureCriterionBase):
    """Defines the core failure criterion."""

    __doc__ = f"""Defines the core shear failure criterion for
    core materials like foam and honeycomb.

    Parameters
    ----------
    include_ins:
        {_DOC_INCLUDE_INS}
    wf:
        {_DOC_WF}
    """

    def __init__(self, *, include_ins: bool = False, wf: float = 1.0):
        """Create a core failure criterion."""
        super().__init__(name="Core Failure", active=True)

        for attr in ATTRS_CORE_FAILURE:
            setattr(self, attr, eval(attr))

    def _get_include_ins(self) -> bool:
        return self._include_ins

    def _set_include_ins(self, value: bool) -> None:
        self._include_ins = value

    def _get_wf(self) -> float:
        return self._wf

    def _set_wf(self, value: float) -> None:
        self._wf = value

    include_ins = property(_get_include_ins, _set_include_ins, doc=_DOC_INCLUDE_INS)
    wf = property(_get_wf, _set_wf, doc=_DOC_WF)


ATTRS_CORE_FAILURE = inspect.signature(CoreFailureCriterion).parameters.keys()



"""Composite Model Interface Factory."""
from typing import Callable, Optional, Union

from ansys.dpf.core import UnitSystem
from ansys.dpf.core.server_types import BaseServer

from ._composite_model_impl import CompositeModelImpl
from ._composite_model_impl_2023r2 import CompositeModelImpl2023R2
from .data_sources import ContinuousFiberCompositesFiles
from .server_helpers import version_older_than

CompositeModelImplT = Callable[
    [ContinuousFiberCompositesFiles, BaseServer, Optional[UnitSystem]],
    Union[CompositeModelImpl2023R2, CompositeModelImpl],
]


def _composite_model_factory(server: BaseServer) -> CompositeModelImplT:
    if version_older_than(server, "7.0"):
        return CompositeModelImpl2023R2

    return CompositeModelImpl



"""Wrapper for the sampling point operator."""
import hashlib
import json
from typing import Any, Collection, Dict, List, Optional, Sequence, Union

import ansys.dpf.core as dpf
from ansys.dpf.core import UnitSystem
from ansys.dpf.core.server import get_or_create_server
from ansys.dpf.core.server_types import BaseServer
import numpy as np
import numpy.typing as npt

from ._sampling_point_helpers import (
    add_ply_sequence_to_plot_to_sp,
    add_results_to_plot_to_sp,
    get_analysis_plies_from_sp,
    get_data_from_sp_results,
    get_indices_from_sp,
    get_offsets_by_spots_from_sp,
    get_ply_wise_critical_failures_from_sp,
    get_polar_plot_from_sp,
    get_result_plots_from_sp,
)
from .constants import Spot
from .result_definition import FailureMeasureEnum, ResultDefinition
from .sampling_point_types import FailureResult, SamplingPoint, SamplingPointFigure
from .server_helpers._load_plugin import load_composites_plugin


class SamplingPoint2023R2(SamplingPoint):
    """Implements the ``Sampling Point`` object that wraps the DPF sampling point operator.

    This class provides for plotting the lay-up and results at a certain point of the
    layered structure. The results, including ``analysis_plies``, ``e1``, ``s12``, and
    ``failure_modes``, are always from the bottom to the top of the laminate (along
    the element normal direction). Postprocessing results such as ``e1`` are returned
    as flat arrays where ``self.spots_per_ply`` can be used to compute the index for
    a certain ply.

    Parameters
    ----------
    name :
        Name of the object.
    result_definition :
        Result definition object that defines all inputs and the scope.

    Notes
    -----
    The results of layered elements are stored per integration point. A layered shell element
    has a number of in-plane integration points (depending on the integration scheme) and
    typically three integration points through the thickness. The through-the-thickness
    integration points are called `spots`. They are typically at the ``BOTTOM``, ``MIDDLE``,
    and ``TOP`` of the layer. This notation is used here to identify the corresponding data.

    The ``SamplingPoint`` class returns three results per layer (one for each spot) because
    the results of the in-plane integration points are interpolated to the centroid of the element.
    The following table shows an example of a laminate with three layers. So a result, such as
    ``s1`` has nine values, three for each ply.

    +------------+------------+------------------------+
    | Layer      | Index      | Spot                   |
    +============+============+========================+
    |            | - 8        | - TOP of Layer 3       |
    | Layer 3    | - 7        | - MIDDLE of Layer 3    |
    |            | - 6        | - BOTTOM of Layer 3    |
    +------------+------------+------------------------+
    |            | - 5        | - TOP of Layer 2       |
    | Layer 2    | - 4        | - MIDDLE of Layer 2    |
    |            | - 3        | - BOTTOM of Layer 2    |
    +------------+------------+------------------------+
    |            | - 2        | - TOP of Layer 1       |
    | Layer 1    | - 1        | - MIDDLE of Layer 1    |
    |            | - 0        | - BOTTOM of Layer 1    |
    +------------+------------+------------------------+

    The get_indices and get_offsets_by_spots methods simplify the indexing and
    filtering of the data.
    """

    def __init__(
        self,
        name: str,
        result_definition: ResultDefinition,
        default_unit_system: Optional[UnitSystem] = None,
        server: BaseServer = None,
    ):
        """Create a ``SamplingPoint`` object."""
        result_definition.check_has_single_scope(f"Sampling point {name} cannot be created.")
        self._result_definition = result_definition

        self._spots_per_ply = 0
        self._interface_indices: Dict[Spot, int] = {}
        self._name = name
        used_server = get_or_create_server(server)
        if not used_server:
            raise RuntimeError("SamplingPoint: cannot connect to DPF server or launch it.")

        if used_server != server:
            load_composites_plugin(used_server)

        self._results: Any = None
        self._is_uptodate = False
        self._rd_hash = ""

        # initialize the sampling point operator. Do it just once
        self._operator = dpf.Operator(
            name="composite::composite_sampling_point_operator", server=used_server
        )
        if not self._operator:
            raise RuntimeError("SamplingPoint: failed to initialize the operator.")
        self._operator.inputs.unit_system(default_unit_system)

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set object name."""
        self._name = value

    @property
    def result_definition(self) -> ResultDefinition:
        """Input for the sampling point operator."""
        return self._result_definition

    @result_definition.setter
    def result_definition(self, value: ResultDefinition) -> None:
        value.check_has_single_scope(f"Cannot set the result definition of {self.name}")
        self._is_uptodate = False
        self._result_definition = value

    @property
    def element_id(self) -> Union[int, None]:
        """Element label for sampling the laminate.

        This attribute returns ``-1`` if the element ID is not set.
        """
        element_scope = self._result_definition.scopes[0].element_scope
        if len(element_scope) > 1:
            raise RuntimeError("The scope of a sampling point can only be one element.")
        if len(element_scope) == 0:
            return None
        return element_scope[0]

    @element_id.setter
    def element_id(self, value: int) -> None:
        self._result_definition.scopes[0].element_scope = [value]
        self._is_uptodate = False

    @property
    def spots_per_ply(self) -> int:
        """Number of through-the-thickness integration points per ply."""
        return self._spots_per_ply

    @property
    def results(self) -> Any:
        """Results of the sampling point operator as a JSON dictionary."""
        self._update_and_check_results()

        return self._results

    @property
    def analysis_plies(self) -> Sequence[Any]:
        """List of analysis plies from the bottom to the top.

        This attribute returns a list of ply data, such as angle, thickness and material name,
        as a dictionary.
        """
        self._update_and_check_results()
        return get_analysis_plies_from_sp(self.results)

    @property
    def s1(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 1 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s1", results=self.results)

    @property
    def s2(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 2 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s2", results=self.results)

    @property
    def s3(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 3 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s3", results=self.results)

    @property
    def s12(self) -> npt.NDArray[np.float64]:
        """In-plane shear stresses s12 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s12", results=self.results)

    @property
    def s13(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear stresses s13 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s13", results=self.results)

    @property
    def s23(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear stresses s23 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s13", results=self.results)

    @property
    def e1(self) -> npt.NDArray[np.float64]:
        """Strains in the material 1 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e1", results=self.results)

    @property
    def e2(self) -> npt.NDArray[np.float64]:
        """Strains in the material 2 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e2", results=self.results)

    @property
    def e3(self) -> npt.NDArray[np.float64]:
        """Strains in the material 3 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e3", results=self.results)

    @property
    def e12(self) -> npt.NDArray[np.float64]:
        """In-plane shear strains e12 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e12", results=self.results)

    @property
    def e13(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear strains e13 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e13", results=self.results)

    @property
    def e23(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear strains e23 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e23", results=self.results)

    @property
    def inverse_reserve_factor(self) -> npt.NDArray[np.float64]:
        """Critical inverse reserve factor of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "inverse_reserve_factor", results=self.results
        )

    @property
    def reserve_factor(self) -> npt.NDArray[np.float64]:
        """Lowest reserve factor of each ply.

        This attribute is equivalent to the safety factor.
        """
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "reserve_factor", results=self.results
        )

    @property
    def margin_of_safety(self) -> npt.NDArray[np.float64]:
        """Lowest margin of safety of each ply.

        This attribute is equivalent to the safety margin.
        """
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "margin_of_safety", results=self.results
        )

    @property
    def failure_modes(self) -> Sequence[str]:
        """Critical failure mode of each ply."""
        self._update_and_check_results()
        return list(self.results[0]["results"]["failures"]["failure_modes"])

    @property
    def offsets(self) -> npt.NDArray[np.float64]:
        """Z coordinates for each interface and ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "offsets", results=self.results)

    @property
    def polar_properties_E1(self) -> npt.NDArray[np.float64]:
        """Polar property E1 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "E1", results=self.results)

    @property
    def polar_properties_E2(self) -> npt.NDArray[np.float64]:
        """Polar property E2 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "E2", results=self.results)

    @property
    def polar_properties_G12(self) -> npt.NDArray[np.float64]:
        """Polar property G12 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "G12", results=self.results)

    @property
    def number_of_plies(self) -> int:
        """Number of plies."""
        return len(self.analysis_plies)

    @property
    def is_uptodate(self) -> bool:
        """True if the results are up-to-date."""
        return self._is_uptodate

    def run(self) -> None:
        """Run the DPF operator and cache the results."""
        if self.result_definition:
            new_hash = hashlib.sha1(
                json.dumps(self.result_definition.to_dict(), sort_keys=True).encode("utf8")
            )
            if new_hash.hexdigest() != self._rd_hash:
                # only set input if the result definition changed
                self._operator.inputs.result_definition(self.result_definition.to_json())
                self._is_uptodate = False
                self._rd_hash = new_hash.hexdigest()
        else:
            raise RuntimeError(
                "Cannot update sampling point because the result definition is missing."
            )

        result_as_string = self._operator.outputs.results()
        self._results = json.loads(result_as_string)
        if not self._results or len(self._results) == 0:
            raise RuntimeError(f"Sampling point {self.name} has no results.")
        if self._results and len(self._results) > 1:
            raise RuntimeError(
                f"Sampling point {self.name} is scoped to more than one element,"
                f" which is not yet supported."
            )

        self._spots_per_ply = 0
        if self._results:
            # update the number of spots
            self._spots_per_ply = int(
                len(np.array(self._results[0]["results"]["strains"]["e1"]))
                / len(self._results[0]["layup"]["analysis_plies"])
            )

        if self._spots_per_ply == 3:
            self._interface_indices = {Spot.BOTTOM: 0, Spot.MIDDLE: 1, Spot.TOP: 2}
        elif self._spots_per_ply == 2:
            self._interface_indices = {Spot.BOTTOM: 0, Spot.TOP: 1}
        elif self._spots_per_ply == 1:
            raise RuntimeError(
                "Result files that only have results at the middle of the ply are not supported."
            )

        self._is_uptodate = True

    def get_indices(
        self, spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP)
    ) -> Sequence[int]:
        """Get the indices of the selected spots (interfaces) for each ply.

        The indices are sorted from bottom to top.
        For instance, this method can be used to access the stresses at the bottom of each ply.

        Parameters
        ----------
        spots :
            Collection of spots. Only the indices of the bottom interfaces of plies
            are returned if ``[Spot.BOTTOM]`` is set.

        Examples
        --------
            >>> ply_top_indices = sampling_point.get_indices([Spot.TOP])

        """
        self._update_and_check_results()
        return get_indices_from_sp(
            self._interface_indices, self.number_of_plies, self.spots_per_ply, spots
        )

    def get_offsets_by_spots(
        self,
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
        core_scale_factor: float = 1.0,
    ) -> npt.NDArray[np.float64]:
        """Access the y coordinates of the selected spots (interfaces) for each ply.

        Parameters
        ----------
        spots :
            Collection of spots.

        core_scale_factor :
            Factor for scaling the thickness of core plies.
        """
        self._update_and_check_results()
        return get_offsets_by_spots_from_sp(self, spots, core_scale_factor)

    def get_ply_wise_critical_failures(self) -> List[FailureResult]:
        """Get the critical failure value and modes per ply."""
        self._update_and_check_results()
        return get_ply_wise_critical_failures_from_sp(self)

    def get_polar_plot(
        self, components: Sequence[str] = ("E1", "E2", "G12")
    ) -> SamplingPointFigure:
        """Create a standard polar plot to visualize the polar properties of the laminate.

        Parameters
        ----------
        components :
            Stiffness quantities to plot.

        Examples
        --------
            >>> figure, axes = sampling_point.get_polar_plot(components=["E1", "G12"])
        """
        self._update_and_check_results()
        return get_polar_plot_from_sp(self, components)

    def add_ply_sequence_to_plot(self, axes: Any, core_scale_factor: float = 1.0) -> None:
        """Add the stacking (ply and text) to an axis or plot.

        Parameters
        ----------
        axes :
            Matplotlib :py:class:`~matplotlib.axes.Axes` object.
        core_scale_factor :
            Factor for scaling the thickness of core plies.
        """
        self._update_and_check_results()
        add_ply_sequence_to_plot_to_sp(self, axes, core_scale_factor)

    def add_results_to_plot(
        self,
        axes: Any,
        components: Sequence[str],
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.TOP),
        core_scale_factor: float = 1.0,
        title: str = "",
        xlabel: str = "",
    ) -> None:
        """Add results (strain, stress, or failure values) to an ``Axes`` object.

        Parameters
        ----------
        axes :
            Matplotlib :py:class:`~matplotlib.axes.Axes` object.
        components :
            List of result components. Valid components for
            strain are ``"e1"``, ``"e2"``, ``"e3"``, ``"e12"``, ``"e13"``,
            and ``"e23"`` Valid components for stress are ``"s1",`` ``"s2"``,
            ``"s3"``, ``"s12"``, ``"s13"``, and ``"s23"``. Valid components
            for failure are ``"inverse_reserve_factor"``, ``"reserve_factor"``,
            and ``"margin_of_safety"``.
        spots :
            Collection of spots (interfaces).
        core_scale_factor :
            Factor for scaling the thickness of core plies.
        title :
            Title of the plot. This parameter is ignored if empty.
        xlabel :
            Becomes the label of the x-axis. This parameter is ignored if empty.

        Examples
        --------
            >>> import matplotlib.pyplot as plt
            >>> fig, ax1 = plt.subplots()
            >>> sampling_point.add_results_to_plot(ax1,
                                                  ["s13", "s23", "s3"],
                                                  [Spot.BOTTOM, Spot.TOP],
                                                  0.1, "Interlaminar Stresses", "[MPa]")
        """
        add_results_to_plot_to_sp(self, axes, components, spots, core_scale_factor, title, xlabel)

    def get_result_plots(
        self,
        strain_components: Sequence[str] = ("e1", "e2", "e3", "e12", "e13", "e23"),
        stress_components: Sequence[str] = ("s1", "s2", "s3", "s12", "s13", "s23"),
        failure_components: Sequence[FailureMeasureEnum] = (
            FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
            FailureMeasureEnum.RESERVE_FACTOR,
            FailureMeasureEnum.MARGIN_OF_SAFETY,
        ),
        show_failure_modes: bool = False,
        create_laminate_plot: bool = True,
        core_scale_factor: float = 1.0,
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
    ) -> SamplingPointFigure:
        """Generate a figure with a grid of axes (plot) for each selected result entity.

        Parameters
        ----------
        strain_components
            Strain entities of interest. Supported values are ``"e1"``, ``"e2"``,
            ``"e3"``, ``"e12"``, ``"e13"``, and ``"e23"``. The plot is skipped
            if the list is empty.
        stress_components
            Stress entities of interest. Supported values are ``"s1"``, ``"s2"``,
            ``"s3"``, ``"s12"``, ``"s13"``, and ``"s23"``. The plot is skipped
            if the list is empty.
        failure_components
            Failure values of interest. Values supported are ``"irf"``, ``"rf"``,
            and ``"mos"``. The plot is skipped if the list is empty.
        show_failure_modes
            WHether to add the critical failure mode to the failure plot.
        create_laminate_plot
            Whether to plot the stacking sequence of the laminate, including text information
            such as material, thickness, and angle.
        core_scale_factor
            Factor for scaling the thickness of core plies.
        spots
            Spots (interfaces) to show results at.

        Examples
        --------
            >>> figure, axes = sampling_point.get_result_plots()

        """
        self._update_and_check_results()
        return get_result_plots_from_sp(
            self,
            strain_components,
            stress_components,
            failure_components,
            show_failure_modes,
            create_laminate_plot,
            core_scale_factor,
            spots,
        )

    def _update_and_check_results(self) -> None:
        if not self._is_uptodate or not self._results:
            self.run()

        if not self._results:
            raise RuntimeError(f"Results of sampling point {self.name} are not available.")



"""Module of failure criteria."""

from ._combined_failure_criterion import CombinedFailureCriterion
from ._core_failure import CoreFailureCriterion
from ._cuntze import CuntzeCriterion
from ._face_sheet_wrinkling import FaceSheetWrinklingCriterion
from ._failure_mode_enum import FailureModeEnum
from ._hashin import HashinCriterion
from ._hoffman import HoffmanCriterion
from ._larc import LaRCCriterion
from ._max_strain import MaxStrainCriterion
from ._max_stress import MaxStressCriterion
from ._puck import PuckCriterion
from ._shear_crimping import ShearCrimpingCriterion
from ._tsai_hill import TsaiHillCriterion
from ._tsai_wu import TsaiWuCriterion
from ._von_mises import VonMisesCriterion

__all__ = [
    "MaxStrainCriterion",
    "MaxStressCriterion",
    "CombinedFailureCriterion",
    "TsaiWuCriterion",
    "TsaiHillCriterion",
    "HoffmanCriterion",
    "CoreFailureCriterion",
    "CuntzeCriterion",
    "FaceSheetWrinklingCriterion",
    "HashinCriterion",
    "LaRCCriterion",
    "PuckCriterion",
    "ShearCrimpingCriterion",
    "VonMisesCriterion",
    "FailureModeEnum",
]



"""Composite Model."""
from typing import Collection, Dict, Optional, Sequence

import ansys.dpf.core as dpf
from ansys.dpf.core import FieldsContainer, MeshedRegion, Operator, UnitSystem
from ansys.dpf.core.server_types import BaseServer
import numpy as np
from numpy.typing import NDArray

from ._composite_model_factory import _composite_model_factory
from .composite_scope import CompositeScope
from .data_sources import CompositeDataSources, ContinuousFiberCompositesFiles
from .failure_criteria import CombinedFailureCriterion
from .layup_info import ElementInfo, LayerProperty
from .layup_info.material_operators import MaterialOperators
from .layup_info.material_properties import MaterialProperty
from .result_definition import FailureMeasureEnum
from .sampling_point_types import SamplingPoint


class CompositeModel:
    """Provides access to the basic composite postprocessing functionality.

    On initialization, the ``CompositeModel`` class automatically adds composite lay-up
    information to the meshed regions. It prepares the providers for different lay-up properties
    so that they can be efficiently evaluated. The composite_files provided are automatically
    uploaded to the server if needed.

    .. note::

        When creating a ``CompositeModel`` instance, several providers are created and
        lay-up information is added to the DPF meshed regions. Depending on the use
        case, it can be more efficient to create the providers separately.

        The handling of models with multiple composite definition files (assemblies)
        differ depending on the version of the DPF server. The handling is simplified
        with DPF Server 7.0 (2024 R1) or later and the full assembly can be post-processed
        in the same way as a model with a single ACP model.

        Before DPF Server 7.0 (2024 R1):

        For assemblies with multiple composite definition files, separate meshes and
        lay-up operators are generated (wrapped by the ``CompositeInfo`` class). This
        is needed because the lay-up provider can only add the data of a single
        composite definitions file to a mesh. All functions that depend on composite
        definitions mut be called with the correct ``composite_definition_label``
        parameter. The layered elements that get information from a given
        composite definition label can be determined by calling
        :meth:`.CompositeModel.get_all_layered_element_ids_for_composite_definition_label`.
        All the elements that are not part of a composite definition are either homogeneous
        solids or layered models defined outside of an ACP model. The
        :meth:`.CompositeModel.composite_definition_labels` command returns all available composite
        definition labels. For more information, see
        :ref:`sphx_glr_examples_gallery_examples_008_assembly_example.py`.


    Parameters
    ----------
    composite_files:
        Use the :func:`.get_composite_files_from_workbench_result_folder` function to obtain
        the :class:`.ContinuousFiberCompositesFiles` object.
    server:
        DPF Server on which the model is created
    default_unit_system:
        Unit system that is used if the result file
        does not specify the unit system. This happens
        for pure MAPDL projects.
    """

    def __init__(
        self,
        composite_files: ContinuousFiberCompositesFiles,
        server: BaseServer,
        default_unit_system: Optional[UnitSystem] = None,
    ):
        """Initialize the composite model class."""
        self._implementation = _composite_model_factory(server)(
            composite_files, server, default_unit_system
        )

    @property
    def composite_definition_labels(self) -> Sequence[str]:
        """All composite definition labels in the model.

        This property is only relevant for assemblies.
        """
        return self._implementation.composite_definition_labels

    @property
    def composite_files(self) -> ContinuousFiberCompositesFiles:
        """Get the composite file paths on the server."""
        return self._implementation.composite_files

    @property
    def data_sources(self) -> CompositeDataSources:
        """Composite data sources."""
        return self._implementation.data_sources

    @property
    def core_model(self) -> dpf.Model:
        """Underlying DPF core model."""
        return self._implementation.core_model

    @property
    def material_operators(self) -> MaterialOperators:
        """Material operators."""
        return self._implementation.material_operators

    def get_mesh(self, composite_definition_label: Optional[str] = None) -> MeshedRegion:
        """Get the underlying DPF meshed region.

        The meshed region contains the lay-up information.

        Parameters
        ----------
        composite_definition_label :
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_mesh(composite_definition_label)

    def get_layup_operator(self, composite_definition_label: Optional[str] = None) -> Operator:
        """Get the lay-up operator.

        Parameters
        ----------
        composite_definition_label :
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.

        """
        return self._implementation.get_layup_operator(composite_definition_label)

    def evaluate_failure_criteria(
        self,
        combined_criterion: CombinedFailureCriterion,
        composite_scope: Optional[CompositeScope] = None,
        measure: FailureMeasureEnum = FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
        write_data_for_full_element_scope: bool = True,
    ) -> FieldsContainer:
        """Get a fields container with the evaluated failure criteria.

        The fields container contains the maximum per element if the measure
        is :attr:`.FailureMeasureEnum.INVERSE_RESERVE_FACTOR` and the minimum per element
        if the measure is :attr:`.FailureMeasureEnum.MARGIN_OF_SAFETY` or
        :attr:`.FailureMeasureEnum.RESERVE_FACTOR`.

        Parameters
        ----------
        combined_criterion :
            Combined failure criterion to evaluate.
        composite_scope :
            Composite scope on which to evaluate the failure criteria. If empty, the criteria
            is evaluated on the full model. If the time is not set, the last time or
            frequency in the result file is used.
        measure :
            Failure measure to evaluate.
        write_data_for_full_element_scope :
            Whether each element in the element scope is to get a
            (potentially zero) failure value, even elements that are not
            part of ``composite_scope.plies``. If no element scope is
            specified (``composite_scope.elements``), a (potentially zero)
            failure value is written for all elements.

            .. note::

                For some special element types such as beams,
                ``write_data_for_full_element_scope=True`` is not supported.

        """
        return self._implementation.evaluate_failure_criteria(
            combined_criterion, composite_scope, measure, write_data_for_full_element_scope
        )

    def get_sampling_point(
        self,
        combined_criterion: CombinedFailureCriterion,
        element_id: int,
        time: Optional[float] = None,
        composite_definition_label: Optional[str] = None,
    ) -> SamplingPoint:
        """Get a sampling point for an element ID and failure criteria.

        Parameters
        ----------
        combined_criterion:
            Combined failure criterion to evaluate.
        element_id:
            Element ID or label of the sampling point.
        time:
            Time or frequency at which to evaluate the sampling point. If ``None``,
            the last time or frequency in the result file is used.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_sampling_point(
            combined_criterion, element_id, time, composite_definition_label
        )

    def get_element_info(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[ElementInfo]:
        """Get element information for an element ID.

        This method returns ``None`` if the element type is not supported.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_element_info(element_id, composite_definition_label)

    def get_property_for_all_layers(
        self,
        layup_property: LayerProperty,
        element_id: int,
        composite_definition_label: Optional[str] = None,
    ) -> Optional[NDArray[np.double]]:
        """Get a layer property for an element ID.

        Returns a numpy array with the values of the property for all the layers.
        Values are ordered from bottom to top.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        layup_property:
            Lay-up property.
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_property_for_all_layers(
            layup_property, element_id, composite_definition_label
        )

    def get_analysis_plies(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[Sequence[str]]:
        """Get analysis ply names.

        This method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the analysis plies in the specified composite
            definition.
        """
        return self._implementation.get_analysis_plies(element_id, composite_definition_label)

    def get_element_laminate_offset(
        self, element_id: int, composite_definition_label: Optional[str] = None
    ) -> Optional[np.double]:
        """Get the laminate offset of an element.

        THis method returns ``None`` if the element is not layered.

        Parameters
        ----------
        element_id:
            Element ID or label.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_element_laminate_offset(
            element_id, composite_definition_label
        )

    def get_constant_property_dict(
        self,
        material_properties: Collection[MaterialProperty],
        composite_definition_label: Optional[str] = None,
    ) -> Dict[np.int64, Dict[MaterialProperty, float]]:
        """Get a dictionary with constant properties.

        Returns a dictionary with ``dpf_material_id`` as the key and
        a dictionary with the requested properties as the value. Only constant properties
        are supported. Variable properties are evaluated at their
        default values.

        This method can be slow to evaluate and should not
        be called in a loop.

        Parameters
        ----------
        material_properties:
            List of the requested material properties.
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            The dictionary only contains the materials of the analysis plies defined
            in the specified composite definition.
        """
        return self._implementation.get_constant_property_dict(
            material_properties, composite_definition_label
        )

    def get_result_times_or_frequencies(self) -> NDArray[np.double]:
        """Get the times or frequencies in the result file."""
        return self._implementation.get_result_times_or_frequencies()

    def add_interlaminar_normal_stresses(
        self,
        stresses: FieldsContainer,
        strains: FieldsContainer,
        composite_definition_label: Optional[str] = None,
    ) -> None:
        """Add interlaminar normal stresses to the stresses fields container.

        Interlaminar normal stresses (s3) are not available for layered shells.
        This function performs a post-processing step which computes s3 and adds
        it to the stress field. s3 is automatically computed if a formulation
        of a failure criterion depends on this stress component, for instance
        :class:`Puck 3D <.failure_criteria.PuckCriterion>` .

        For a usage example, see
        :ref:`sphx_glr_examples_gallery_examples_007_interlaminar_normal_stress_example.py`.

        Parameters
        ----------
        stresses:
            Stresses fields container to add interlaminar normal stresses to.
        strains:
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
            Interlaminar normal stresses are only added to the layered elements defined
            in the specified composite definition.
        """
        self._implementation.add_interlaminar_normal_stresses(
            stresses, strains, composite_definition_label
        )

    def get_all_layered_element_ids(self) -> Sequence[int]:
        """Get all element IDs with lay-up data."""
        return self._implementation.get_all_layered_element_ids()

    def get_all_layered_element_ids_for_composite_definition_label(
        self, composite_definition_label: Optional[str] = None
    ) -> Sequence[int]:
        """Get all layered element IDs that belong to a composite definition label.

        Parameters
        ----------
        composite_definition_label:
            Label of the composite definition, which is the
            dictionary key in the :attr:`.ContinuousFiberCompositesFiles.composite`
            attribute. This parameter is only required for assemblies.
            See the note about assemblies in the description for the :class:`CompositeModel` class.
        """
        return self._implementation.get_all_layered_element_ids_for_composite_definition_label(
            composite_definition_label
        )



"""Von Mises Criterion."""
import inspect

from ._failure_criterion_base import FailureCriterionBase

_DOC_VME = "Whether to activate the von Mises strain criterion."
_DOC_VMS = "Whether to activate the von Mises stress criterion."
_DOC_WF_VME = "Weighting factor of the strain criterion (vme)."
_DOC_WF_VMS = "Weighting factor of the stress criterion (vms)."
_DOC_EVAL_INS = (
    "Whether to consider interlaminar normal stresses (s3) and compute them for layered shells."
)


class VonMisesCriterion(FailureCriterionBase):
    """Von Mises criterion."""

    __doc__ = f"""Defines the von Mises criterion for isotropic materials.

    Parameters
    ----------
    vme:
        {_DOC_VME}
    vms:
        {_DOC_VMS}
    wf_vme:
        {_DOC_WF_VME}
    wf_vms:
        {_DOC_WF_VMS}
    eval_ins:
        {_DOC_EVAL_INS}
    """

    def __init__(
        self,
        *,
        vme: bool = True,
        vms: bool = True,
        wf_vme: float = 1.0,
        wf_vms: float = 1.0,
        eval_ins: bool = False,
    ):
        """Create a von Mises criterion."""
        super().__init__(name="Von Mises", active=True)

        for attr in ATTRS_VON_MISES:
            setattr(self, attr, eval(attr))

    def _get_vme(self) -> bool:
        return self._vme

    def _set_vme(self, value: bool) -> None:
        self._vme = value

    def _get_vms(self) -> bool:
        return self._vms

    def _set_vms(self, value: bool) -> None:
        self._vms = value

    def _get_wf_vme(self) -> float:
        return self._wf_vme

    def _set_wf_vme(self, value: float) -> None:
        self._wf_vme = value

    def _get_wf_vms(self) -> float:
        return self._wf_vms

    def _set_wf_vms(self, value: float) -> None:
        self._wf_vms = value

    def _get_eval_ins(self) -> bool:
        return self._eval_ins

    def _set_eval_ins(self, value: bool) -> None:
        self._eval_ins = value

    vme = property(_get_vme, _set_vme, doc=_DOC_VME)
    vms = property(_get_vms, _set_vms, doc=_DOC_VMS)
    wf_vme = property(_get_wf_vme, _set_wf_vme, doc=_DOC_WF_VME)
    wf_vms = property(_get_wf_vms, _set_wf_vms, doc=_DOC_WF_VMS)

    eval_ins = property(
        _get_eval_ins,
        _set_eval_ins,
        doc=_DOC_EVAL_INS,
    )


ATTRS_VON_MISES = inspect.signature(VonMisesCriterion).parameters.keys()



"""Wrapper for the sampling point operator."""
import json
from typing import Any, Collection, Dict, List, Optional, Sequence, Union

from ansys.dpf.core import UnitSystem
import numpy as np
import numpy.typing as npt

from ansys.dpf import core as dpf

from ._sampling_point_helpers import (
    add_ply_sequence_to_plot_to_sp,
    add_results_to_plot_to_sp,
    get_analysis_plies_from_sp,
    get_data_from_sp_results,
    get_indices_from_sp,
    get_offsets_by_spots_from_sp,
    get_ply_wise_critical_failures_from_sp,
    get_polar_plot_from_sp,
    get_result_plots_from_sp,
)
from .constants import Spot
from .failure_criteria import CombinedFailureCriterion
from .layup_info.material_operators import MaterialOperators
from .result_definition import FailureMeasureEnum
from .sampling_point_types import FailureResult, SamplingPoint, SamplingPointFigure
from .unit_system import get_unit_system


class SamplingPointNew(SamplingPoint):
    """Implements the ``Sampling Point`` object that wraps the DPF sampling point operator.

    This class provides for plotting the lay-up and results at a certain point of the
    layered structure. The results, including ``analysis_plies``, ``e1``, ``s12``, and
    ``failure_modes``, are always from the bottom to the top of the laminate (along
    the element normal direction). Postprocessing results such as ``e1`` are returned
    as flat arrays where ``self.spots_per_ply`` can be used to compute the index for
    a certain ply.

    Parameters
    ----------
    name :
        Name of the object.

    Notes
    -----
    The results of layered elements are stored per integration point. A layered shell element
    has a number of in-plane integration points (depending on the integration scheme) and
    typically three integration points through the thickness. The through-the-thickness
    integration points are called `spots`. They are typically at the ``BOTTOM``, ``MIDDLE``,
    and ``TOP`` of the layer. This notation is used here to identify the corresponding data.

    The ``SamplingPoint`` class returns three results per layer (one for each spot) because
    the results of the in-plane integration points are interpolated to the centroid of the element.
    The following table shows an example of a laminate with three layers. So a result, such as
    ``s1`` has nine values, three for each ply.

    +------------+------------+------------------------+
    | Layer      | Index      | Spot                   |
    +============+============+========================+
    |            | - 8        | - TOP of Layer 3       |
    | Layer 3    | - 7        | - MIDDLE of Layer 3    |
    |            | - 6        | - BOTTOM of Layer 3    |
    +------------+------------+------------------------+
    |            | - 5        | - TOP of Layer 2       |
    | Layer 2    | - 4        | - MIDDLE of Layer 2    |
    |            | - 3        | - BOTTOM of Layer 2    |
    +------------+------------+------------------------+
    |            | - 2        | - TOP of Layer 1       |
    | Layer 1    | - 1        | - MIDDLE of Layer 1    |
    |            | - 0        | - BOTTOM of Layer 1    |
    +------------+------------+------------------------+

    The get_indices and get_offsets_by_spots methods simplify the indexing and
    filtering of the data.
    """

    def __init__(
        self,
        name: str,
        element_id: int,
        combined_criterion: CombinedFailureCriterion,
        material_operators: MaterialOperators,
        meshed_region: dpf.MeshedRegion,
        layup_provider: dpf.Operator,
        rst_streams_provider: dpf.Operator,
        rst_data_source: dpf.DataSources,
        default_unit_system: Optional[UnitSystem] = None,
        time: Optional[float] = None,
    ):
        """Create a ``SamplingPoint`` object."""
        self._name = name
        self._element_id = element_id
        self._time = time
        self._combined_criterion = combined_criterion

        self._material_operators = material_operators
        self._meshed_region = meshed_region
        self._layup_provider = layup_provider
        self._rst_streams_provider = rst_streams_provider
        self._rst_data_source = rst_data_source

        self._spots_per_ply = 0
        self._interface_indices: Dict[Spot, int] = {}
        self._results: Any = None
        self._is_uptodate = False
        self._unit_system = get_unit_system(self._rst_data_source, default_unit_system)

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set object name."""
        self._name = value

    @property
    def element_id(self) -> Union[int, None]:
        """Element label for sampling the laminate."""
        return self._element_id

    @element_id.setter
    def element_id(self, value: int) -> None:
        self._element_id = value
        self._is_uptodate = False

    @property
    def combined_criterion(self) -> CombinedFailureCriterion:
        """Element label for sampling the laminate."""
        return self._combined_criterion

    @combined_criterion.setter
    def combined_criterion(self, value: CombinedFailureCriterion) -> None:
        self._combined_criterion = value
        self._is_uptodate = False

    @property
    def spots_per_ply(self) -> int:
        """Number of through-the-thickness integration points per ply."""
        return self._spots_per_ply

    @property
    def results(self) -> Any:
        """Results of the sampling point operator as a JSON dictionary."""
        self._update_and_check_results()
        return self._results

    @property
    def analysis_plies(self) -> Sequence[Any]:
        """List of analysis plies from the bottom to the top.

        This attribute returns a list of ply data, such as angle, thickness and material name,
        as a dictionary.
        """
        self._update_and_check_results()
        return get_analysis_plies_from_sp(self.results)

    @property
    def s1(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 1 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s1", results=self.results)

    @property
    def s2(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 2 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s2", results=self.results)

    @property
    def s3(self) -> npt.NDArray[np.float64]:
        """Stresses in the material 3 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s3", results=self.results)

    @property
    def s12(self) -> npt.NDArray[np.float64]:
        """In-plane shear stresses s12 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s12", results=self.results)

    @property
    def s13(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear stresses s13 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s13", results=self.results)

    @property
    def s23(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear stresses s23 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "stresses", "s13", results=self.results)

    @property
    def e1(self) -> npt.NDArray[np.float64]:
        """Strains in the material 1 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e1", results=self.results)

    @property
    def e2(self) -> npt.NDArray[np.float64]:
        """Strains in the material 2 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e2", results=self.results)

    @property
    def e3(self) -> npt.NDArray[np.float64]:
        """Strains in the material 3 direction of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e3", results=self.results)

    @property
    def e12(self) -> npt.NDArray[np.float64]:
        """In-plane shear strains e12 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e12", results=self.results)

    @property
    def e13(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear strains e13 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e13", results=self.results)

    @property
    def e23(self) -> npt.NDArray[np.float64]:
        """Out-of-plane shear strains e23 of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "strains", "e23", results=self.results)

    @property
    def inverse_reserve_factor(self) -> npt.NDArray[np.float64]:
        """Critical inverse reserve factor of each ply."""
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "inverse_reserve_factor", results=self.results
        )

    @property
    def reserve_factor(self) -> npt.NDArray[np.float64]:
        """Lowest reserve factor of each ply.

        This attribute is equivalent to the safety factor.
        """
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "reserve_factor", results=self.results
        )

    @property
    def margin_of_safety(self) -> npt.NDArray[np.float64]:
        """Lowest margin of safety of each ply.

        This attribute is equivalent to the safety margin.
        """
        self._update_and_check_results()
        return get_data_from_sp_results(
            "results", "failures", "margin_of_safety", results=self.results
        )

    @property
    def failure_modes(self) -> Sequence[str]:
        """Critical failure mode of each ply."""
        self._update_and_check_results()
        return list(self.results[0]["results"]["failures"]["failure_modes"])

    @property
    def offsets(self) -> npt.NDArray[np.float64]:
        """Z coordinates for each interface and ply."""
        self._update_and_check_results()
        return get_data_from_sp_results("results", "offsets", results=self.results)

    @property
    def polar_properties_E1(self) -> npt.NDArray[np.float64]:
        """Polar property E1 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "E1", results=self.results)

    @property
    def polar_properties_E2(self) -> npt.NDArray[np.float64]:
        """Polar property E2 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "E2", results=self.results)

    @property
    def polar_properties_G12(self) -> npt.NDArray[np.float64]:
        """Polar property G12 of the laminate."""
        self._update_and_check_results()
        return get_data_from_sp_results("layup", "polar_properties", "G12", results=self.results)

    @property
    def number_of_plies(self) -> int:
        """Number of plies."""
        return len(self.analysis_plies)

    @property
    def is_uptodate(self) -> bool:
        """True if the results are up-to-date."""
        return self._is_uptodate

    def run(self) -> None:
        """Build and run the DPF operator network and cache the results."""
        scope_config_reader_op = dpf.Operator("composite::scope_config_reader")
        scope_config = dpf.DataTree()
        if self._time:
            scope_config.add({"requested_times": [self._time]})
        scope_config_reader_op.inputs.scope_configuration(scope_config)

        evaluate_failure_criterion_per_scope_op = dpf.Operator(
            "composite::evaluate_failure_criterion_per_scope"
        )

        # Live evaluation is not yet supported
        evaluate_failure_criterion_per_scope_op.inputs.criterion_configuration(
            self.combined_criterion.to_json()
        )

        evaluate_failure_criterion_per_scope_op.inputs.scope_configuration(
            scope_config_reader_op.outputs
        )

        scope = dpf.Scoping()
        scope.ids = [self.element_id]
        evaluate_failure_criterion_per_scope_op.inputs.element_scoping(scope)
        evaluate_failure_criterion_per_scope_op.inputs.materials_container(
            self._material_operators.material_provider.outputs
        )
        evaluate_failure_criterion_per_scope_op.inputs.stream_provider(self._rst_streams_provider)
        evaluate_failure_criterion_per_scope_op.inputs.mesh(self._meshed_region)
        has_layup_provider = True
        evaluate_failure_criterion_per_scope_op.inputs.has_layup_provider(has_layup_provider)
        evaluate_failure_criterion_per_scope_op.inputs.section_data_container(
            self._layup_provider.outputs.section_data_container
        )

        evaluate_failure_criterion_per_scope_op.inputs.material_fields(
            self._layup_provider.outputs.material_fields
        )
        evaluate_failure_criterion_per_scope_op.inputs.mesh_properties_container(
            self._layup_provider.outputs.mesh_properties_container
        )
        evaluate_failure_criterion_per_scope_op.inputs.request_sandwich_results(True)

        sampling_point_evaluator = dpf.Operator("composite::evaluate_sampling_point")

        sampling_point_evaluator.inputs.materials_container(
            self._material_operators.material_provider.outputs
        )
        sampling_point_evaluator.inputs.material_support(
            self._material_operators.material_support_provider.outputs
        )
        sampling_point_evaluator.inputs.mesh(self._meshed_region)
        sampling_point_evaluator.inputs.stresses_container(
            evaluate_failure_criterion_per_scope_op.outputs.stresses_container
        )
        sampling_point_evaluator.inputs.strains_container(
            evaluate_failure_criterion_per_scope_op.outputs.strains_container
        )
        sampling_point_evaluator.inputs.element_scoping(scope)
        sampling_point_evaluator.inputs.section_data_container(
            self._layup_provider.outputs.section_data_container
        )

        sampling_point_evaluator.inputs.time_id(
            evaluate_failure_criterion_per_scope_op.outputs.time_id
        )
        sampling_point_evaluator.inputs.unit_system(self._unit_system)
        sampling_point_evaluator.inputs.failure_container(
            evaluate_failure_criterion_per_scope_op.outputs.failure_container
        )
        sampling_point_evaluator.inputs.extract_max_failure_per_layer(False)
        sampling_point_evaluator.inputs.check_mechanical_unit_system(False)

        sampling_point_to_json_converter = dpf.Operator("composite::convert_sampling_point_to_json")
        sampling_point_to_json_converter.connect(0, sampling_point_evaluator, 0)

        # update internal members
        self._results = json.loads(
            sampling_point_to_json_converter.get_output(pin=0, output_type=dpf.types.string)
        )
        if not self._results or len(self._results) == 0:
            raise RuntimeError(f"Sampling point {self.name} has no results.")
        if self._results and len(self._results) > 1:
            raise RuntimeError(
                f"Sampling point {self.name} is scoped to more than one element,"
                f" which is not yet supported."
            )

        self._spots_per_ply = 0
        if self._results:
            # update the number of spots
            self._spots_per_ply = int(
                len(np.array(self._results[0]["results"]["strains"]["e1"]))
                / len(self._results[0]["layup"]["analysis_plies"])
            )

        if self._spots_per_ply == 3:
            self._interface_indices = {Spot.BOTTOM: 0, Spot.MIDDLE: 1, Spot.TOP: 2}
        elif self._spots_per_ply == 2:
            self._interface_indices = {Spot.BOTTOM: 0, Spot.TOP: 1}
        elif self._spots_per_ply == 1:
            raise RuntimeError(
                "Result files that only have results at the middle of the ply are not supported."
            )

        self._is_uptodate = True

    def get_indices(
        self, spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP)
    ) -> Sequence[int]:
        """Get the indices of the selected spots (interfaces) for each ply.

        The indices are sorted from bottom to top.
        For instance, this method can be used to access the stresses at the bottom of each ply.

        Parameters
        ----------
        spots :
            Collection of spots. Only the indices of the bottom interfaces of plies
            are returned if ``[Spot.BOTTOM]`` is set.

        Examples
        --------
            >>> ply_top_indices = sampling_point.get_indices([Spot.TOP])

        """
        self._update_and_check_results()
        return get_indices_from_sp(
            self._interface_indices, self.number_of_plies, self.spots_per_ply, spots
        )

    def get_offsets_by_spots(
        self,
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
        core_scale_factor: float = 1.0,
    ) -> npt.NDArray[np.float64]:
        """Access the y coordinates of the selected spots (interfaces) for each ply.

        Parameters
        ----------
        spots :
            Collection of spots.

        core_scale_factor :
            Factor for scaling the thickness of core plies.
        """
        self._update_and_check_results()
        return get_offsets_by_spots_from_sp(self, spots, core_scale_factor)

    def get_ply_wise_critical_failures(self) -> List[FailureResult]:
        """Get the critical failure value and modes per ply."""
        self._update_and_check_results()
        return get_ply_wise_critical_failures_from_sp(self)

    def add_results_to_plot(
        self,
        axes: Any,
        components: Sequence[str],
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.TOP),
        core_scale_factor: float = 1.0,
        title: str = "",
        xlabel: str = "",
    ) -> None:
        """Add results (strain, stress, or failure values) to an ``Axes`` object.

        Parameters
        ----------
        axes :
            Matplotlib :py:class:`~matplotlib.axes.Axes` object.
        components :
            List of result components. Valid components for
            strain are ``"e1"``, ``"e2"``, ``"e3"``, ``"e12"``, ``"e13"``,
            and ``"e23"`` Valid components for stress are ``"s1",`` ``"s2"``,
            ``"s3"``, ``"s12"``, ``"s13"``, and ``"s23"``. Valid components
            for failure are ``"inverse_reserve_factor"``, ``"reserve_factor"``,
            and ``"margin_of_safety"``.
        spots :
            Collection of spots (interfaces).
        core_scale_factor :
            Factor for scaling the thickness of core plies.
        title :
            Title of the plot. This parameter is ignored if empty.
        xlabel :
            Becomes the label of the x-axis. This parameter is ignored if empty.

        Examples
        --------
            >>> import matplotlib.pyplot as plt
            >>> fig, ax1 = plt.subplots()
            >>> sampling_point.add_results_to_plot(ax1,
                                                  ["s13", "s23", "s3"],
                                                  [Spot.BOTTOM, Spot.TOP],
                                                  0.1, "Interlaminar Stresses", "[MPa]")
        """
        self._update_and_check_results()
        add_results_to_plot_to_sp(self, axes, components, spots, core_scale_factor, title, xlabel)

    def add_ply_sequence_to_plot(self, axes: Any, core_scale_factor: float = 1.0) -> None:
        """Add the stacking (ply and text) to an axis or plot.

        Parameters
        ----------
        axes :
            Matplotlib :py:class:`~matplotlib.axes.Axes` object.
        core_scale_factor :
            Factor for scaling the thickness of core plies.
        """
        self._update_and_check_results()
        add_ply_sequence_to_plot_to_sp(self, axes, core_scale_factor)

    def get_polar_plot(
        self, components: Sequence[str] = ("E1", "E2", "G12")
    ) -> SamplingPointFigure:
        """Create a standard polar plot to visualize the polar properties of the laminate.

        Parameters
        ----------
        components :
            Stiffness quantities to plot.

        Examples
        --------
            >>> figure, axes = sampling_point.get_polar_plot(components=["E1", "G12"])
        """
        self._update_and_check_results()
        return get_polar_plot_from_sp(self, components)

    def get_result_plots(
        self,
        strain_components: Sequence[str] = ("e1", "e2", "e3", "e12", "e13", "e23"),
        stress_components: Sequence[str] = ("s1", "s2", "s3", "s12", "s13", "s23"),
        failure_components: Sequence[FailureMeasureEnum] = (
            FailureMeasureEnum.INVERSE_RESERVE_FACTOR,
            FailureMeasureEnum.RESERVE_FACTOR,
            FailureMeasureEnum.MARGIN_OF_SAFETY,
        ),
        show_failure_modes: bool = False,
        create_laminate_plot: bool = True,
        core_scale_factor: float = 1.0,
        spots: Collection[Spot] = (Spot.BOTTOM, Spot.MIDDLE, Spot.TOP),
    ) -> SamplingPointFigure:
        """Generate a figure with a grid of axes (plot) for each selected result entity.

        Parameters
        ----------
        strain_components
            Strain entities of interest. Supported values are ``"e1"``, ``"e2"``,
            ``"e3"``, ``"e12"``, ``"e13"``, and ``"e23"``. The plot is skipped
            if the list is empty.
        stress_components
            Stress entities of interest. Supported values are ``"s1"``, ``"s2"``,
            ``"s3"``, ``"s12"``, ``"s13"``, and ``"s23"``. The plot is skipped
            if the list is empty.
        failure_components
            Failure values of interest. Values supported are ``"irf"``, ``"rf"``,
            and ``"mos"``. The plot is skipped if the list is empty.
        show_failure_modes
            WHether to add the critical failure mode to the failure plot.
        create_laminate_plot
            Whether to plot the stacking sequence of the laminate, including text information
            such as material, thickness, and angle.
        core_scale_factor
            Factor for scaling the thickness of core plies.
        spots
            Spots (interfaces) to show results at.

        Examples
        --------
            >>> figure, axes = sampling_point.get_result_plots()

        """
        self._update_and_check_results()
        return get_result_plots_from_sp(
            self,
            strain_components,
            stress_components,
            failure_components,
            show_failure_modes,
            create_laminate_plot,
            core_scale_factor,
            spots,
        )

    def _update_and_check_results(self) -> None:
        if not self._is_uptodate or not self._results:
            self.run()

        if not self._results:
            raise RuntimeError(f"Results of sampling point {self.name} are not available.")



"""Base class for quadratic failure criterion."""

from ._failure_criterion_base import FailureCriterionBase

_DOC_WF = "Weighting factor of this failure criterion."
_DOC_DIM = "Specifies which formulation of the failure criterion is used."


class QuadraticFailureCriterion(FailureCriterionBase):
    """Base class for quadratic failure criteria.

    Such as Tsai-Wu, Tsai-Hill and Hoffman.
    """

    def __init__(self, *, name: str, active: bool, wf: float, dim: int):
        """Do not use directly. Base class for quadratic failure criteria."""
        super().__init__(name=name, active=active)

        self.dim = dim
        self.wf = wf

    def _get_wf(self) -> float:
        return self._wf

    def _set_wf(self, value: float) -> None:
        self._wf = value

    def _get_dim(self) -> int:
        return self._dim

    def _set_dim(self, value: int) -> None:
        if value in [2, 3]:
            self._dim = value
        else:
            raise AttributeError(
                f"Dimension of {self.name} cannot be set to {value}. Allowed are 2 or 3."
            )

    wf = property(_get_wf, _set_wf, doc=_DOC_WF)
    dim = property(_get_dim, _set_dim, doc=_DOC_DIM)
